"""This module contains types defined by this package.

These are all simple types using named tuples so can be safely ignored by downstream users provided data
fields are passed in the correct order.
"""

import dataclasses
from datetime import timedelta

import polars as pl

# The type used for final aggregate counts of predicates.
PRED_CNT_TYPE = pl.Int64

# The key used in the endpoint expression to indicate the window should be aggregated to the record start.
START_OF_RECORD_KEY = "_RECORD_START"
END_OF_RECORD_KEY = "_RECORD_END"

# The key used to capture the count of events of any kind that occur in a window.
ANY_EVENT_COLUMN = "_ANY_EVENT"


@dataclasses.dataclass(order=True)
class TemporalWindowBounds:
    """Named tuple to represent temporal window bounds.

    Attributes:
        left_inclusive: The start of the window, inclusive.
        window_size: The size of the window.
        right_inclusive: The end of the window, inclusive.
        offset: The offset from the start of the window to the end of the window.

    Example:
        >>> bounds = TemporalWindowBounds(
        ...     left_inclusive=True,
        ...     window_size=timedelta(days=1),
        ...     right_inclusive=False,
        ...     offset=timedelta(hours=1)
        ... )
        >>> bounds # doctest: +NORMALIZE_WHITESPACE
        TemporalWindowBounds(left_inclusive=True,
                             window_size=datetime.timedelta(days=1),
                             right_inclusive=False,
                             offset=datetime.timedelta(seconds=3600))
        >>> left_inclusive, window_size, right_inclusive, offset = bounds
        >>> bounds.left_inclusive
        True
        >>> window_size
        datetime.timedelta(days=1)
        >>> right_inclusive
        False
        >>> offset
        datetime.timedelta(seconds=3600)
    """

    left_inclusive: bool
    window_size: timedelta
    right_inclusive: bool
    offset: timedelta | None = None

    # Needed to make it accessible like a tuple.
    def __iter__(self):
        return (getattr(self, field.name) for field in dataclasses.fields(self))

    # Needed to make it scriptable.
    def __getitem__(self, key):
        return tuple(getattr(self, field.name) for field in dataclasses.fields(self))[key]

    def __post_init__(self):
        if self.offset is None:
            self.offset = timedelta(0)

    @property
    def polars_gp_rolling_kwargs(self) -> dict[str, str | timedelta]:
        """Return the parameters for a group_by rolling operation in Polars.

        Examples:
            >>> TemporalWindowBounds(
            ...     left_inclusive=True,
            ...     window_size=timedelta(days=1),
            ...     right_inclusive=True,
            ...     offset=None
            ... ).polars_gp_rolling_kwargs # doctest: +NORMALIZE_WHITESPACE
            {'period': datetime.timedelta(days=1),
             'offset': datetime.timedelta(0),
             'closed': 'both'}
            >>> TemporalWindowBounds(
            ...     left_inclusive=True,
            ...     window_size=timedelta(days=1),
            ...     right_inclusive=True,
            ...     offset=timedelta(hours=1)
            ... ).polars_gp_rolling_kwargs # doctest: +NORMALIZE_WHITESPACE
            {'period': datetime.timedelta(days=1),
             'offset': datetime.timedelta(seconds=3600),
             'closed': 'both'}
            >>> TemporalWindowBounds(
            ...     left_inclusive=False,
            ...     window_size=timedelta(days=2),
            ...     right_inclusive=False,
            ...     offset=timedelta(minutes=1)
            ... ).polars_gp_rolling_kwargs # doctest: +NORMALIZE_WHITESPACE
            {'period': datetime.timedelta(days=2),
             'offset': datetime.timedelta(seconds=60),
             'closed': 'none'}
            >>> TemporalWindowBounds(
            ...     left_inclusive=True,
            ...     window_size=timedelta(days=2),
            ...     right_inclusive=False,
            ...     offset=timedelta(minutes=1)
            ... ).polars_gp_rolling_kwargs # doctest: +NORMALIZE_WHITESPACE
            {'period': datetime.timedelta(days=2),
             'offset': datetime.timedelta(seconds=60),
             'closed': 'left'}
            >>> TemporalWindowBounds(
            ...     left_inclusive=False,
            ...     window_size=timedelta(days=2),
            ...     right_inclusive=True,
            ...     offset=timedelta(minutes=1)
            ... ).polars_gp_rolling_kwargs # doctest: +NORMALIZE_WHITESPACE
            {'period': datetime.timedelta(days=2),
             'offset': datetime.timedelta(seconds=60),
             'closed': 'right'}
        """
        if self.left_inclusive and self.right_inclusive:
            closed = "both"
        elif self.left_inclusive:
            closed = "left"
        elif self.right_inclusive:
            closed = "right"
        else:
            closed = "none"

        # set parameters for group_by rolling window
        if self.window_size < timedelta(days=0):
            period = -self.window_size
            offset = -period + self.offset
        else:
            period = self.window_size
            offset = self.offset

        return {"period": period, "offset": offset, "closed": closed}


@dataclasses.dataclass(order=True)
class ToEventWindowBounds:
    """Named tuple to represent temporal window bounds.

    Attributes:
        left_inclusive: The start of the window, inclusive.
        end_event: The string name of the event that bounds the end of this window. Operationally, this is
            interpreted as the string name of the column which contains a positive value if the row
            corresponds to the end event of this window and a zero otherwise.
        right_inclusive: The end of the window, inclusive.
        offset: The offset from the start of the window to the end of the window.

    Raises:
        ValueError: If `end_event` is an empty string.
        ValueError: If `offset` is negative.

    Example:
        >>> bounds = ToEventWindowBounds(
        ...     left_inclusive=True,
        ...     end_event="foo",
        ...     right_inclusive=False,
        ...     offset=timedelta(hours=1)
        ... )
        >>> bounds # doctest: +NORMALIZE_WHITESPACE
        ToEventWindowBounds(left_inclusive=True,
                            end_event='foo',
                            right_inclusive=False,
                            offset=datetime.timedelta(seconds=3600))
        >>> left_inclusive, end_event, right_inclusive, offset = bounds
        >>> left_inclusive
        True
        >>> end_event
        'foo'
        >>> right_inclusive
        False
        >>> offset
        datetime.timedelta(seconds=3600)
        >>> bounds = ToEventWindowBounds(
        ...     left_inclusive=True,
        ...     end_event="",
        ...     right_inclusive=False,
        ...     offset=timedelta(hours=1)
        ... )
        Traceback (most recent call last):
            ...
        ValueError: The 'end_event' must be a non-empty string.
        >>> bounds = ToEventWindowBounds(
        ...     left_inclusive=True,
        ...     end_event="_RECORD_START",
        ...     right_inclusive=False,
        ...     offset=timedelta(hours=1)
        ... ) # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        ValueError: It doesn't make sense to have the start of the record _RECORD_START be an end event. Did
        you mean to make that be the start event (which should result in the `end_event` parameter being
        '-_RECORD_START'?
    """

    left_inclusive: bool
    end_event: str
    right_inclusive: bool
    offset: timedelta | None = None

    def __post_init__(self):
        if self.end_event == "":
            raise ValueError("The 'end_event' must be a non-empty string.")

        if self.end_event == START_OF_RECORD_KEY:
            raise ValueError(
                f"It doesn't make sense to have the start of the record {START_OF_RECORD_KEY} be an end "
                "event. Did you mean to make that be the start event (which should result in the `end_event` "
                f"parameter being '-{START_OF_RECORD_KEY}'?"
            )
        elif self.end_event == f"-{END_OF_RECORD_KEY}":
            raise ValueError(
                f"It doesn't make sense to have the end of the record {END_OF_RECORD_KEY} be a start "
                "event. Did you mean to make that be the end event (which should result in the `end_event` "
                f"parameter being '{END_OF_RECORD_KEY}'?"
            )

        if self.offset is None:
            self.offset = timedelta(0)

    # Needed to make it accessible like a tuple.
    def __iter__(self):
        return (getattr(self, field.name) for field in dataclasses.fields(self))

    # Needed to make it scriptable.
    def __getitem__(self, key):
        return tuple(getattr(self, field.name) for field in dataclasses.fields(self))[key]

    @property
    def boolean_expr_bound_sum_kwargs(self) -> dict[str, str | timedelta | pl.Expr]:
        """Return the parameters for a group_by rolling operation in Polars.

        Examples:
            >>> def print_kwargs(kwargs: dict):
            ...     for key, value in kwargs.items():
            ...         print(f"{key}: {value}")
            >>> print_kwargs(ToEventWindowBounds(
            ...     left_inclusive=True, end_event="is_A", right_inclusive=False, offset=None
            ... ).boolean_expr_bound_sum_kwargs) # doctest: +NORMALIZE_WHITESPACE
            boundary_expr: [(col("is_A")) > (dyn int: 0)]
            mode: row_to_bound
            closed: left
            offset: 0:00:00
            >>> print_kwargs(ToEventWindowBounds(
            ...     left_inclusive=False, end_event="-is_B", right_inclusive=True, offset=None
            ... ).boolean_expr_bound_sum_kwargs) # doctest: +NORMALIZE_WHITESPACE
            boundary_expr: [(col("is_B")) > (dyn int: 0)]
            mode: bound_to_row
            closed: right
            offset: 0:00:00
            >>> print_kwargs(ToEventWindowBounds(
            ...     left_inclusive=False, end_event="is_B", right_inclusive=False, offset=timedelta(hours=-3)
            ... ).boolean_expr_bound_sum_kwargs) # doctest: +NORMALIZE_WHITESPACE
            boundary_expr: [(col("is_B")) > (dyn int: 0)]
            mode: row_to_bound
            closed: none
            offset: -1 day, 21:00:00
            >>> print_kwargs(ToEventWindowBounds(
            ...     left_inclusive=True,
            ...     end_event="-_RECORD_START",
            ...     right_inclusive=True,
            ...     offset=timedelta(days=2),
            ... ).boolean_expr_bound_sum_kwargs) # doctest: +NORMALIZE_WHITESPACE
            boundary_expr: [(col("timestamp")) == (col("timestamp").min().over([col("subject_id")]))]
            mode: bound_to_row
            closed: both
            offset: 2 days, 0:00:00
            >>> print_kwargs(ToEventWindowBounds(
            ...     left_inclusive=False,
            ...     end_event="_RECORD_END",
            ...     right_inclusive=True,
            ...     offset=timedelta(days=1),
            ... ).boolean_expr_bound_sum_kwargs) # doctest: +NORMALIZE_WHITESPACE
            boundary_expr: [(col("timestamp")) == (col("timestamp").max().over([col("subject_id")]))]
            mode: row_to_bound
            closed: right
            offset: 1 day, 0:00:00
        """

        if self.left_inclusive and self.right_inclusive:
            closed = "both"
        elif (not self.left_inclusive) and (not self.right_inclusive):
            closed = "none"
        elif self.left_inclusive:
            closed = "left"
        elif self.right_inclusive:
            closed = "right"

        mode = "bound_to_row" if self.end_event.startswith("-") else "row_to_bound"

        if mode == "bound_to_row":
            end_event = self.end_event[1:]
        else:
            end_event = self.end_event

        if end_event == START_OF_RECORD_KEY:
            boundary_expr = pl.col("timestamp") == pl.col("timestamp").min().over("subject_id")
        elif end_event == END_OF_RECORD_KEY:
            boundary_expr = pl.col("timestamp") == pl.col("timestamp").max().over("subject_id")
        else:
            boundary_expr = pl.col(end_event) > 0

        return {
            "boundary_expr": boundary_expr,
            "mode": mode,
            "closed": closed,
            "offset": self.offset,
        }



"""Contains utilities for validating that windows satisfy a set of constraints."""

import polars as pl
from loguru import logger

from .types import ANY_EVENT_COLUMN


def check_constraints(
    window_constraints: dict[str, tuple[int | None, int | None]], summary_df: pl.DataFrame
) -> pl.DataFrame:
    """Checks the constraints on the counts of predicates in the summary dataframe.

    Args:
        window_constraints: constraints on counts of predicates that must
            be satsified, organized as a dictionary from predicate column name to the lowerbound and upper
            bound range required for that constraint to be satisfied.
        summary_df: A dataframe containing a row for every possible prospective window to be analyzed. The
            only columns expected are predicate columns within the ``window_constraints`` dictionary.

    Returns: A filtered dataframe containing only the rows that satisfy the constraints.

    Raises:
        ValueError: If the constraint for a column is empty.

    Examples:
        >>> from datetime import datetime
        >>> df = pl.DataFrame({
        ...     "subject_id": [1, 1, 1, 1, 2, 2],
        ...     "timestamp": [
        ...         # Subject 1
        ...         datetime(year=1989, month=12, day=1, hour=12, minute=3),
        ...         datetime(year=1989, month=12, day=2, hour=5,  minute=17),
        ...         datetime(year=1989, month=12, day=2, hour=12, minute=3),
        ...         datetime(year=1989, month=12, day=6, hour=11, minute=0),
        ...         # Subject 2
        ...         datetime(year=1989, month=12, day=1, hour=13, minute=14),
        ...         datetime(year=1989, month=12, day=3, hour=15, minute=17),
        ...     ],
        ...     "is_A": [1, 4, 1, 3, 3,  3],
        ...     "is_B": [0, 2, 0, 2, 10, 2],
        ...     "is_C": [1, 1, 1, 0, 1,  1],
        ... })
        >>> check_constraints({"is_A": (None, None), "is_B": (2, 6), "is_C": (1, 1)}, df)
        Traceback (most recent call last):
            ...
        ValueError: Invalid constraint for 'is_A': None - None
        >>> check_constraints({"is_A": (2, 1), "is_B": (2, 6), "is_C": (1, 1)}, df)
        Traceback (most recent call last):
            ...
        ValueError: Invalid constraint for 'is_A': 2 - 1
        >>> check_constraints({"is_A": (3, 4), "is_B": (2, 6), "is_C": (1, 1)}, df)
        shape: (2, 5)
        ┌────────────┬─────────────────────┬──────┬──────┬──────┐
        │ subject_id ┆ timestamp           ┆ is_A ┆ is_B ┆ is_C │
        │ ---        ┆ ---                 ┆ ---  ┆ ---  ┆ ---  │
        │ i64        ┆ datetime[μs]        ┆ i64  ┆ i64  ┆ i64  │
        ╞════════════╪═════════════════════╪══════╪══════╪══════╡
        │ 1          ┆ 1989-12-02 05:17:00 ┆ 4    ┆ 2    ┆ 1    │
        │ 2          ┆ 1989-12-03 15:17:00 ┆ 3    ┆ 2    ┆ 1    │
        └────────────┴─────────────────────┴──────┴──────┴──────┘
        >>> check_constraints({"is_A": (3, 4), "is_B": (2, None), "is_C": (None, 1)}, df)
        shape: (4, 5)
        ┌────────────┬─────────────────────┬──────┬──────┬──────┐
        │ subject_id ┆ timestamp           ┆ is_A ┆ is_B ┆ is_C │
        │ ---        ┆ ---                 ┆ ---  ┆ ---  ┆ ---  │
        │ i64        ┆ datetime[μs]        ┆ i64  ┆ i64  ┆ i64  │
        ╞════════════╪═════════════════════╪══════╪══════╪══════╡
        │ 1          ┆ 1989-12-02 05:17:00 ┆ 4    ┆ 2    ┆ 1    │
        │ 1          ┆ 1989-12-06 11:00:00 ┆ 3    ┆ 2    ┆ 0    │
        │ 2          ┆ 1989-12-01 13:14:00 ┆ 3    ┆ 10   ┆ 1    │
        │ 2          ┆ 1989-12-03 15:17:00 ┆ 3    ┆ 2    ┆ 1    │
        └────────────┴─────────────────────┴──────┴──────┴──────┘
        >>> predicates_df = pl.DataFrame({
        ...     "subject_id": [1, 1, 3],
        ...     "timestamp": [datetime(1980, 12, 28), datetime(2010, 6, 20), datetime(2010, 5, 11)],
        ...     "A": [False, False, False],
        ...     "_ANY_EVENT": [True, True, True],
        ... })
        >>> check_constraints({"_ANY_EVENT": (1, None)}, predicates_df)
        shape: (3, 4)
        ┌────────────┬─────────────────────┬───────┬────────────┐
        │ subject_id ┆ timestamp           ┆ A     ┆ _ANY_EVENT │
        │ ---        ┆ ---                 ┆ ---   ┆ ---        │
        │ i64        ┆ datetime[μs]        ┆ bool  ┆ bool       │
        ╞════════════╪═════════════════════╪═══════╪════════════╡
        │ 1          ┆ 1980-12-28 00:00:00 ┆ false ┆ true       │
        │ 1          ┆ 2010-06-20 00:00:00 ┆ false ┆ true       │
        │ 3          ┆ 2010-05-11 00:00:00 ┆ false ┆ true       │
        └────────────┴─────────────────────┴───────┴────────────┘
    """

    should_drop = pl.lit(False)

    for col, (valid_min_inc, valid_max_inc) in window_constraints.items():
        if valid_min_inc is None and valid_max_inc is None:
            raise ValueError(f"Invalid constraint for '{col}': {valid_min_inc} - {valid_max_inc}")
        elif valid_min_inc is not None and valid_max_inc is not None and valid_max_inc < valid_min_inc:
            raise ValueError(f"Invalid constraint for '{col}': {valid_min_inc} - {valid_max_inc}")

        if col == "*":
            col = ANY_EVENT_COLUMN

        drop_expr = pl.lit(False)
        if valid_min_inc is not None:
            drop_expr = drop_expr | (pl.col(col) < valid_min_inc)
        if valid_max_inc is not None:
            drop_expr = drop_expr | (pl.col(col) > valid_max_inc)

        logger.info(
            f"Excluding {summary_df.select(drop_expr.sum()).item():,} rows "
            f"as they failed to satisfy '{valid_min_inc} <= {col} <= {valid_max_inc}'."
        )

        should_drop = should_drop | drop_expr

    return summary_df.filter(~should_drop)


def check_static_variables(patient_demographics: list[str], predicates_df: pl.DataFrame) -> pl.DataFrame:
    """Checks the constraints on the counts of predicates in the summary dataframe.

    Args:
        patient_demographics: List of columns representing static patient demographics.
        predicates_df: Dataframe containing a row for each event with patient demographics and timestamps.

    Returns: A filtered dataframe containing only the rows that satisfy the patient demographics.

    Raises:
        ValueError: If the static predicate used by constraint is not in the predicates dataframe.

    Examples:
        >>> from datetime import datetime
        >>> predicates_df = pl.DataFrame({
        ...     "subject_id": [1, 1, 1, 1, 1, 2, 2, 2],
        ...     "timestamp": [
        ...         # Subject 1
        ...         None,
        ...         datetime(year=1989, month=12, day=1, hour=12, minute=3),
        ...         datetime(year=1989, month=12, day=2, hour=5,  minute=17),
        ...         datetime(year=1989, month=12, day=2, hour=12, minute=3),
        ...         datetime(year=1989, month=12, day=6, hour=11, minute=0),
        ...         # Subject 2
        ...         None,
        ...         datetime(year=1989, month=12, day=1, hour=13, minute=14),
        ...         datetime(year=1989, month=12, day=3, hour=15, minute=17),
        ...     ],
        ...     "is_A": [0, 1, 4, 1, 0, 3, 3,  3],
        ...     "is_B": [0, 0, 2, 0, 0, 2, 10, 2],
        ...     "is_C": [0, 1, 1, 1, 0, 0, 1,  1],
        ...     "male": [1, 0, 0, 0, 0, 0, 0,  0]
        ... })

        >>> check_static_variables(['male'], predicates_df)
        shape: (4, 5)
        ┌────────────┬─────────────────────┬──────┬──────┬──────┐
        │ subject_id ┆ timestamp           ┆ is_A ┆ is_B ┆ is_C │
        │ ---        ┆ ---                 ┆ ---  ┆ ---  ┆ ---  │
        │ i64        ┆ datetime[μs]        ┆ i64  ┆ i64  ┆ i64  │
        ╞════════════╪═════════════════════╪══════╪══════╪══════╡
        │ 1          ┆ 1989-12-01 12:03:00 ┆ 1    ┆ 0    ┆ 1    │
        │ 1          ┆ 1989-12-02 05:17:00 ┆ 4    ┆ 2    ┆ 1    │
        │ 1          ┆ 1989-12-02 12:03:00 ┆ 1    ┆ 0    ┆ 1    │
        │ 1          ┆ 1989-12-06 11:00:00 ┆ 0    ┆ 0    ┆ 0    │
        └────────────┴─────────────────────┴──────┴──────┴──────┘
    """
    for demographic in patient_demographics:
        if demographic not in predicates_df.columns:
            raise ValueError(f"Static predicate '{demographic}' not found in the predicates dataframe.")

        keep_expr = ((pl.col("timestamp").is_null()) & (pl.col(demographic) == 1)).alias("keep_expr")

        exclude_expr = ~keep_expr
        exclude_count = predicates_df.filter(exclude_expr).shape[0]

        logger.info(f"Excluding {exclude_count:,} rows due to the '{demographic}' criteria.")

        predicates_df = predicates_df.filter(
            pl.col("subject_id").is_in(predicates_df.filter(keep_expr).select("subject_id").unique())
        )

    return predicates_df.drop_nulls(subset=["timestamp"]).drop(
        *[x for x in patient_demographics if x in predicates_df.columns]
    )



"""Test utilities."""

import rootutils

root = rootutils.setup_root(__file__, dotenv=True, pythonpath=True, cwd=True)

import subprocess
import tempfile
from pathlib import Path

import polars as pl
from loguru import logger
from polars.testing import assert_frame_equal


def run_command(script: str, hydra_kwargs: dict[str, str], test_name: str, expected_returncode: int = 0):
    command_parts = [script] + [f"{k}={v}" for k, v in hydra_kwargs.items()]
    cmd = " ".join(command_parts)
    logger.info(f"Running {test_name}: {cmd}")
    command_out = subprocess.run(cmd, shell=True, capture_output=True)
    stderr = command_out.stderr.decode()
    stdout = command_out.stdout.decode()
    if command_out.returncode != expected_returncode:
        raise AssertionError(
            f"{test_name} returned {command_out.returncode} (expected {expected_returncode})!\n"
            f"stdout:\n{stdout}\nstderr:\n{stderr}"
        )
    return stderr, stdout


def assert_df_equal(want: pl.DataFrame, got: pl.DataFrame, msg: str = None, **kwargs):
    try:
        assert_frame_equal(want, got, **kwargs)
    except AssertionError as e:
        pl.Config.set_tbl_rows(-1)
        print(f"DFs are not equal: {msg}\nWant:")
        print(want)
        print("Got:")
        print(got)
        raise AssertionError(f"{msg}\n{e}") from e


def write_input_files(data_dir: Path, input_files: dict[str, pl.DataFrame | str]) -> dict[str, Path]:
    wrote_files = {}
    for name, df in input_files.items():
        if isinstance(df, str):
            data_fp = data_dir / f"{name}.csv"
            data_fp.parent.mkdir(parents=True, exist_ok=True)
            data_fp.write_text(df)
        elif isinstance(df, pl.DataFrame):
            data_fp = data_dir / f"{name}.parquet"
            data_fp.parent.mkdir(parents=True, exist_ok=True)
            df.write_parquet(data_fp)
        else:
            raise ValueError(f"Invalid input type: {type(df)}")
        wrote_files[name] = data_fp
    return wrote_files


def write_task_configs(
    cohort_dir: Path, task_configs: dict[str, str], predicate_files: dict[str, str] | None = None
) -> dict[str, Path]:
    wrote_files = {}
    for name, cfg in task_configs.items():
        fp = cohort_dir / f"{name}.yaml"
        fp.parent.mkdir(parents=True, exist_ok=True)
        fp.write_text(cfg)
        wrote_files[name] = fp
    if predicate_files is not None:
        for name, cfg in predicate_files.items():
            fp = cohort_dir / f"{name}_predicates.yaml"
            fp.parent.mkdir(parents=True, exist_ok=True)
            fp.write_text(cfg)
            wrote_files[f"{name}_predicates"] = fp
    return wrote_files


def cli_test(
    input_files: dict[str, pl.DataFrame | str],
    task_configs: dict[str, str],
    want_outputs_by_task: dict[str, dict[str, pl.DataFrame]],
    data_standard: str,
    predicate_files: str = None,
):
    if data_standard not in ["meds", "direct"]:
        raise ValueError(f"Invalid data standard: {data_standard}")

    with tempfile.TemporaryDirectory() as root_dir:
        root_dir = Path(root_dir)
        data_dir = root_dir / "sample_data" / "data"
        cohort_dir = root_dir / "sample_cohort"

        wrote_files = write_input_files(data_dir, input_files)
        if len(wrote_files) == 0:
            raise ValueError("No input files were written.")
        elif len(wrote_files) > 1:
            sharded = True
            command = "aces-cli --multirun"
        else:
            sharded = False
            command = "aces-cli"

        wrote_configs = write_task_configs(cohort_dir, task_configs, predicate_files)
        if len(wrote_configs) == 0:
            raise ValueError("No task configs were written.")

        for task in task_configs:
            if sharded:
                want_outputs = {
                    cohort_dir / task / f"{n}.parquet": df for n, df in want_outputs_by_task[task].items()
                }
            else:
                want_outputs = {cohort_dir / f"{task}.parquet": want_outputs_by_task[task]}

            extraction_config_kwargs = {
                "cohort_dir": str(cohort_dir.resolve()),
                "cohort_name": task,
                "hydra.verbose": True,
                "data.standard": data_standard,
            }

            if len(wrote_files) > 1:
                extraction_config_kwargs["data"] = "sharded"
                extraction_config_kwargs["data.root"] = str(data_dir.resolve())
                extraction_config_kwargs['"data.shard'] = f'$(expand_shards {str(data_dir.resolve())})"'
            else:
                extraction_config_kwargs["data.path"] = str(list(wrote_files.values())[0].resolve())

            if predicate_files is not None:
                extraction_config_kwargs["predicates_path"] = str(
                    wrote_configs[f"{task}_predicates"].resolve()
                )

            stderr, stdout = run_command(command, extraction_config_kwargs, f"CLI should run for {task}")

            try:
                if sharded:
                    out_dir = cohort_dir / task
                    all_out_fps = list(out_dir.glob("**/*.parquet"))
                    all_out_fps_str = ", ".join(str(x.relative_to(out_dir)) for x in all_out_fps)
                    if len(all_out_fps) == 0 and len(want_outputs) > 0:
                        all_directory_contents = ", ".join(
                            str(x.relative_to(cohort_dir)) for x in cohort_dir.glob("**/*")
                        )

                        raise AssertionError(
                            f"No output files found for task '{task}'. Found files: {all_directory_contents}"
                        )

                    assert len(all_out_fps) == len(
                        want_outputs
                    ), f"Expected {len(want_outputs)} outputs, got {len(all_out_fps)}: {all_out_fps_str}"

                for want_fp, want_df in want_outputs.items():
                    out_shard = want_fp.relative_to(cohort_dir)
                    assert want_fp.is_file(), f"Expected {out_shard} to exist."

                    got_df = pl.read_parquet(want_fp)
                    assert_df_equal(
                        want_df, got_df, f"Data mismatch for shard '{out_shard}':\n{want_df}\n{got_df}"
                    )
            except AssertionError as e:
                logger.error(f"{stderr}\n{stdout}")
                raise AssertionError(f"Error running task '{task}': {e}") from e



"""This module contains functions for loading and parsing the configuration file and subsequently building a
tree structure from the configuration."""

from __future__ import annotations

import dataclasses
import re
from collections import OrderedDict
from dataclasses import field
from datetime import timedelta
from pathlib import Path

import networkx as nx
import polars as pl
import ruamel.yaml
from bigtree import Node
from loguru import logger

from .types import (
    ANY_EVENT_COLUMN,
    END_OF_RECORD_KEY,
    START_OF_RECORD_KEY,
    TemporalWindowBounds,
    ToEventWindowBounds,
)
from .utils import parse_timedelta


@dataclasses.dataclass
class PlainPredicateConfig:
    code: str | dict
    value_min: float | None = None
    value_max: float | None = None
    value_min_inclusive: bool | None = None
    value_max_inclusive: bool | None = None
    static: bool = False
    other_cols: dict[str, str] = field(default_factory=dict)

    def MEDS_eval_expr(self) -> pl.Expr:
        """Returns a Polars expression that evaluates this predicate for a MEDS formatted dataset.

        Note: The output syntax for the following examples is dependent on the polars version used. The
        expected outputs have been validated on polars version 0.20.30.

        Examples:
            >>> expr = PlainPredicateConfig("BP//systolic", 120, 140, True, False).MEDS_eval_expr()
            >>> print(expr) # doctest: +NORMALIZE_WHITESPACE
            [(col("code")) == (String(BP//systolic))].all_horizontal([[(col("numerical_value")) >=
               (dyn int: 120)], [(col("numerical_value")) < (dyn int: 140)]])
            >>> cfg = PlainPredicateConfig("BP//systolic", value_min=120, value_min_inclusive=False)
            >>> expr = cfg.MEDS_eval_expr()
            >>> print(expr) # doctest: +NORMALIZE_WHITESPACE
            [(col("code")) == (String(BP//systolic))].all_horizontal([[(col("numerical_value")) >
               (dyn int: 120)]])
            >>> cfg = PlainPredicateConfig("BP//systolic", value_max=140, value_max_inclusive=True)
            >>> expr = cfg.MEDS_eval_expr()
            >>> print(expr) # doctest: +NORMALIZE_WHITESPACE
            [(col("code")) == (String(BP//systolic))].all_horizontal([[(col("numerical_value")) <=
               (dyn int: 140)]])
            >>> cfg = PlainPredicateConfig("BP//diastolic")
            >>> expr = cfg.MEDS_eval_expr()
            >>> print(expr) # doctest: +NORMALIZE_WHITESPACE
            [(col("code")) == (String(BP//diastolic))]
            >>> cfg = PlainPredicateConfig("BP//diastolic", other_cols={"chamber": "atrial"})
            >>> expr = cfg.MEDS_eval_expr()
            >>> print(expr) # doctest: +NORMALIZE_WHITESPACE
            [(col("code")) == (String(BP//diastolic))].all_horizontal([[(col("chamber")) ==
               (String(atrial))]])

            >>> cfg = PlainPredicateConfig(code={'regex': None, 'any': None})
            >>> expr = cfg.MEDS_eval_expr() # doctest: +NORMALIZE_WHITESPACE
            Traceback (most recent call last):
                ...
            ValueError: Only one of 'regex' or 'any' can be specified in the code field!
            Got: ['regex', 'any'].
            >>> cfg = PlainPredicateConfig(code={'foo': None})
            >>> expr = cfg.MEDS_eval_expr() # doctest: +NORMALIZE_WHITESPACE
            Traceback (most recent call last):
                ...
            ValueError: Invalid specification in the code field! Got: {'foo': None}.
            Expected one of 'regex', 'any'.
            >>> cfg = PlainPredicateConfig(code={'regex': ''})
            >>> expr = cfg.MEDS_eval_expr() # doctest: +NORMALIZE_WHITESPACE
            Traceback (most recent call last):
                ...
            ValueError: Invalid specification in the code field! Got: {'regex': ''}.
            Expected a non-empty string for 'regex'.
            >>> cfg = PlainPredicateConfig(code={'any': []})
            >>> expr = cfg.MEDS_eval_expr() # doctest: +NORMALIZE_WHITESPACE
            Traceback (most recent call last):
                ...
            ValueError: Invalid specification in the code field! Got: {'any': []}.
            Expected a list of strings for 'any'.

            >>> cfg = PlainPredicateConfig(code={'regex': '^foo.*'})
            >>> expr = cfg.MEDS_eval_expr() # doctest: +NORMALIZE_WHITESPACE
            >>> print(expr) # doctest: +NORMALIZE_WHITESPACE
            col("code").str.contains([String(^foo.*)])
            >>> cfg = PlainPredicateConfig(code={'any': ['foo', 'bar']})
            >>> expr = cfg.MEDS_eval_expr() # doctest: +NORMALIZE_WHITESPACE
            >>> print(expr) # doctest: +NORMALIZE_WHITESPACE
            col("code").is_in([Series])
        """
        criteria = []
        if isinstance(self.code, dict):
            if len(self.code) > 1:
                raise ValueError(
                    "Only one of 'regex' or 'any' can be specified in the code field! "
                    f"Got: {list(self.code.keys())}."
                )

            if "regex" in self.code:
                if not self.code["regex"] or not isinstance(self.code["regex"], str):
                    raise ValueError(
                        "Invalid specification in the code field! "
                        f"Got: {self.code}. "
                        "Expected a non-empty string for 'regex'."
                    )
                criteria.append(pl.col("code").str.contains(self.code["regex"]))
            elif "any" in self.code:
                if not self.code["any"] or not isinstance(self.code["any"], list):
                    raise ValueError(
                        "Invalid specification in the code field! "
                        f"Got: {self.code}. "
                        f"Expected a list of strings for 'any'."
                    )
                criteria.append(pl.Expr.is_in(pl.col("code"), self.code["any"]))
            else:
                raise ValueError(
                    "Invalid specification in the code field! "
                    f"Got: {self.code}. "
                    "Expected one of 'regex', 'any'."
                )
        else:
            criteria.append(pl.col("code") == self.code)

            if self.value_min is not None:
                if self.value_min_inclusive:
                    criteria.append(pl.col("numerical_value") >= self.value_min)
                else:
                    criteria.append(pl.col("numerical_value") > self.value_min)
            if self.value_max is not None:
                if self.value_max_inclusive:
                    criteria.append(pl.col("numerical_value") <= self.value_max)
                else:
                    criteria.append(pl.col("numerical_value") < self.value_max)

        if self.other_cols:
            criteria.extend([pl.col(col) == value for col, value in self.other_cols.items()])

        if len(criteria) == 1:
            return criteria[0]
        else:
            return pl.all_horizontal(criteria)

    def ESGPT_eval_expr(self, values_column: str | None = None) -> pl.Expr:
        """Returns a Polars expression that evaluates this predicate for a ESGPT formatted dataset.

        Note: The output syntax for the following examples is dependent on the polars version used. The
        expected outputs have been validated on polars version 0.20.30.

        Examples:
            >>> # Should handle univariate regression values
            >>> expr = PlainPredicateConfig("HR", value_min=120, value_min_inclusive=False)
            >>> expr = expr.ESGPT_eval_expr("HR")
            >>> print(expr)
            [(col("HR")) > (dyn int: 120)]
            >>> expr = PlainPredicateConfig("BP//systolic", 120, 140, True, False).ESGPT_eval_expr("BP_value")
            >>> print(expr) # doctest: +NORMALIZE_WHITESPACE
            [(col("BP")) == (String(systolic))].all_horizontal([[(col("BP_value")) >=
               (dyn int: 120)], [(col("BP_value")) < (dyn int: 140)]])
            >>> cfg = PlainPredicateConfig("BP//systolic", value_min=120, value_min_inclusive=False)
            >>> expr = cfg.ESGPT_eval_expr("blood_pressure_value")
            >>> print(expr) # doctest: +NORMALIZE_WHITESPACE
            [(col("BP")) == (String(systolic))].all_horizontal([[(col("blood_pressure_value")) >
               (dyn int: 120)]])
            >>> cfg = PlainPredicateConfig("BP//systolic", value_max=140, value_max_inclusive=True)
            >>> expr = cfg.ESGPT_eval_expr("blood_pressure_value")
            >>> print(expr) # doctest: +NORMALIZE_WHITESPACE
            [(col("BP")) == (String(systolic))].all_horizontal([[(col("blood_pressure_value")) <=
               (dyn int: 140)]])
            >>> expr = PlainPredicateConfig("BP//diastolic").ESGPT_eval_expr()
            >>> print(expr) # doctest: +NORMALIZE_WHITESPACE
            [(col("BP")) == (String(diastolic))]
            >>> expr = PlainPredicateConfig("event_type//ADMISSION").ESGPT_eval_expr()
            >>> print(expr) # doctest: +NORMALIZE_WHITESPACE
            col("event_type").strict_cast(String).str.split([String(&)]).list.contains([String(ADMISSION)])
            >>> expr = PlainPredicateConfig("BP//diastolic//atrial").ESGPT_eval_expr()
            >>> print(expr) # doctest: +NORMALIZE_WHITESPACE
            [(col("BP")) == (String(diastolic//atrial))]
            >>> expr = PlainPredicateConfig("BP//diastolic", None, None).ESGPT_eval_expr()
            >>> print(expr) # doctest: +NORMALIZE_WHITESPACE
            [(col("BP")) == (String(diastolic))]
            >>> expr = PlainPredicateConfig("BP").ESGPT_eval_expr()
            >>> print(expr) # doctest: +NORMALIZE_WHITESPACE
            col("BP").is_not_null()
            >>> expr = PlainPredicateConfig("BP//systole", other_cols={"chamber": "atrial"}).ESGPT_eval_expr()
            >>> print(expr) # doctest: +NORMALIZE_WHITESPACE
            [(col("BP")) == (String(systole))].all_horizontal([[(col("chamber")) == (String(atrial))]])
            >>> expr = PlainPredicateConfig("BP//systolic", value_min=120).ESGPT_eval_expr()
            Traceback (most recent call last):
                ...
            ValueError: Must specify a values column for ESGPT predicates with a value_min = 120
            >>> expr = PlainPredicateConfig("BP//systolic", value_max=140).ESGPT_eval_expr()
            Traceback (most recent call last):
                ...
            ValueError: Must specify a values column for ESGPT predicates with a value_max = 140
        """
        code_is_in_parts = "//" in self.code

        if code_is_in_parts:
            codes = self.code.split("//")
            measurement_name = codes.pop(0)
            if len(codes) > 1:
                code = "//".join(codes)
            else:
                code = codes[0]
            if measurement_name.lower() == "event_type":
                criteria = [pl.col("event_type").cast(pl.Utf8).str.split("&").list.contains(code)]
            else:
                criteria = [pl.col(measurement_name) == code]
        elif (self.value_min is None) and (self.value_max is None):
            return pl.col(self.code).is_not_null()
        else:
            values_column = self.code
            criteria = []

        if self.value_min is not None:
            if values_column is None:
                raise ValueError(
                    f"Must specify a values column for ESGPT predicates with a value_min = {self.value_min}"
                )
            if self.value_min_inclusive:
                criteria.append(pl.col(values_column) >= self.value_min)
            else:
                criteria.append(pl.col(values_column) > self.value_min)
        if self.value_max is not None:
            if values_column is None:
                raise ValueError(
                    f"Must specify a values column for ESGPT predicates with a value_max = {self.value_max}"
                )
            if self.value_max_inclusive:
                criteria.append(pl.col(values_column) <= self.value_max)
            else:
                criteria.append(pl.col(values_column) < self.value_max)

        if self.other_cols:
            criteria.extend([pl.col(col) == value for col, value in self.other_cols.items()])

        if len(criteria) == 1:
            return criteria[0]
        else:
            return pl.all_horizontal(criteria)

    @property
    def is_plain(self) -> bool:
        return True


@dataclasses.dataclass
class DerivedPredicateConfig:
    """A configuration object for derived predicates, which are composed of multiple input predicates.

    Args:
        expr: The expression defining the derived predicate in terms of other predicates.

    Raises:
        ValueError: If the expression is empty, does not start with 'and(' or 'or(', or does not contain at
            least two input predicates.

    Examples:
        >>> pred = DerivedPredicateConfig("and(P1, P2, P3)")
        >>> pred = DerivedPredicateConfig("and()") # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        ValueError: Derived predicate expression must have at least two input predicates (comma separated).
        Got: 'and()'
        >>> pred = DerivedPredicateConfig("or(PA, PB)")
        >>> pred = DerivedPredicateConfig("PA + PB")
        Traceback (most recent call last):
            ...
        ValueError: Derived predicate expression must start with 'and(' or 'or('. Got: 'PA + PB'
        >>> pred = DerivedPredicateConfig("")
        Traceback (most recent call last):
            ...
        ValueError: Derived predicates must have a non-empty expression field.
    """

    expr: str
    static: bool = False

    def __post_init__(self):
        if not self.expr:
            raise ValueError("Derived predicates must have a non-empty expression field.")

        self.is_and = self.expr.startswith("and(") and self.expr.endswith(")")
        self.is_or = self.expr.startswith("or(") and self.expr.endswith(")")
        if not (self.is_and or self.is_or):
            raise ValueError(
                f"Derived predicate expression must start with 'and(' or 'or('. Got: '{self.expr}'"
            )

        if self.is_and:
            self.input_predicates = [x.strip() for x in self.expr[4:-1].split(",")]
        elif self.is_or:
            self.input_predicates = [x.strip() for x in self.expr[3:-1].split(",")]

        if len(self.input_predicates) < 2:
            raise ValueError(
                "Derived predicate expression must have at least two input predicates (comma separated). "
                f"Got: '{self.expr}'"
            )

    def eval_expr(self) -> pl.Expr:
        """Returns a Polars expression that evaluates this predicate against necessary dependent predicates.

        Note: The output syntax for the following examples is dependent on the polars version used. The
        expected outputs have been validated on polars version 0.20.30.

        Examples:
            >>> expr = DerivedPredicateConfig("and(P1, P2, P3)").eval_expr()
            >>> print(expr) # doctest: +NORMALIZE_WHITESPACE
            [(col("P1")) > (dyn int: 0)].all_horizontal([[(col("P2")) >
               (dyn int: 0)], [(col("P3")) > (dyn int: 0)]])
            >>> expr = DerivedPredicateConfig("or(PA, PB)").eval_expr()
            >>> print(expr)
            [(col("PA")) > (dyn int: 0)].any_horizontal([[(col("PB")) > (dyn int: 0)]])
        """
        if self.is_and:
            return pl.all_horizontal([pl.col(pred) > 0 for pred in self.input_predicates])
        elif self.is_or:
            return pl.any_horizontal([pl.col(pred) > 0 for pred in self.input_predicates])

    @property
    def is_plain(self) -> bool:
        return False


@dataclasses.dataclass
class WindowConfig:
    """A configuration object for defining a window in the task extraction process.

    This defines the boundary points and constraints for a window in the patient record in the task extraction
    process.

    Args:
        start: The boundary conditions for the start of the window. This (like ``end``) can either be `None`,
            in which case the window starts at the beginning of the patient record, or is expressed through a
            string language that expresses a relative startpoint to this window either in reference to (a)
            another window's start or end event, (b) this window's `end` event. In case (a), this window's end
            event must either be `None` or reference this window's start event, and in case (b), this window's
            end event must reference a different window's start or end event.
            The string language is as follows:
              - ``None``: The window starts at the beginning of the patient record.
              - ``$REFERENCED <- $PREDICATE`` or ``$REFERENCED -> $PREDICATE``: The window starts at the
                closest event satisfying the predicate ``$PREDICATE`` relative to the ``$REFERENCED`` event.
                Form ``$REFERENCED <- $PREDICATE`` means that the window starts at the closest event _prior
                to_ the ``$REFERENCED`` event that satisfies the predicate ``$PREDICATE``, and the other form
                is analogous but with the closest event _after_ the ``$REFERENCED`` event.
              - ``$REFERENCED +- timedelta``: The window starts at the ``$REFERENCED`` event plus or minus the
                specified timedelta. The timedelta is expressed through the string language specified in the
                `utils.parse_timedelta` function.
              - ``$REFERENCED``: The window starts at the ``$REFERENCED`` event.
            In all cases, the ``$REFERENCED`` event must be either
              - The name of another window's start or end event, as specified by ``$WINDOW_NAME.start`` or
                ``$WINDOW_NAME.end``.
              - This window's end event, as specified by ``end``.
            In the case that ``$REFERENCED`` is this window's end event, the window must be defined such that
            ``start`` would precede ``end`` in the order of the patient record (e.g., ``$PREDICATE -> end`` is
            invalid, and ``end - timedelta`` is invalid).
        end: The name of the event that ends the window. See the documentation for ``start`` for more details
            on the specification language.
        start_inclusive: Whether or not the start event is included in the window. Note that this term can not
            only dictate whether an event's counts are included in the summarization of the window, but also
            whether or not an event satisfying ``$PREDICATE`` can be used as the boundary of an event. E.g.,
            if we have that `start_inclusive=False` and the `end` field is equal to `start -> $PREDICATE`, and
            it so happens that the `start` event itself satisfies `$PREDICATE`, the fact that
            `start_inclusive=False` will mean that we do not consider the `start` event itself to be a valid
            start to any window that ends at the same `start` event, as its timestamp when considered as the
            prospective "window start timestamp" occurs "after" the effective timestamp of itself when
            considered as the `$PREDICATE` event that marks the window end given that `start_inclusive=False`
            and thus we will think of the window as truly starting an iota after the timestamp of the `start`
            event itself.
        end_inclusive: Whether or not the end event is included in the window.
        has: A dictionary of predicates that must be present in the window, mapped to tuples of the form
            `(min_valid, max_valid)` that define the valid range the count of observations of the named
            predicate that must be found in a window for it to be considered valid. Either `min_valid` or
            `max_valid` constraints can be `None`, in which case those endpoints are left unconstrained.
            Likewise, unreferenced predicates are also left unconstrained. Note that as predicate counts are
            always integral, this specification does not need an additional inclusive/exclusive endpoint
            field, as one can simply increment the bound by one in the appropriate direction to achieve the
            result. Instead, this bound is always interpreted to be inclusive, so a window would satisfy the
            constraint for predicate `name` with constraint `name: (1, 2)` if the count of observations of
            predicate `name` in a window was either 1 or 2. All constraints in the dictionary must be
            satisfied on a window for it to be included.
        label: A string that specifies the name of a predicate to be used as the label for the task. The
            predicate count of the window this field is specified in will be extracted as a column in the
            final result. Hence, there can only be one 'label' per TaskExtractorConfig. If more than one
            'label' is specified, an error is raised. If the specified 'label' is not a defined predicate,
            an error is also raised. If no 'label' is specified, there will be not be a 'label' column.
        index_timestamp: A string that is either 'start' or 'end' and is used to index result rows. If it is
            defined, there will be an 'index_timestamp' column in the result with its values equal to the
            'start' or 'end' timestamp of the window in which it was specified. Usually, this will be
            specified to indicate the time of prediction for the task, which is often the 'end' of the input
            window. There can only be one 'index_timestamp' per TaskExtractorConfig. If more than one
            'index_timestamp' is specified, an error is raised. If the specified 'index_timestamp' is not
            'start' or 'end', an error is also raised. If no 'index_timestamp' is defined, there will be no
            'index_timestamp' column.

    Raises:
        ValueError: If the window is misconfigured in any of a variety of ways; see below for examples.

    Examples:
        >>> input_window = WindowConfig(
        ...     start=None,
        ...     end="trigger + 2 days",
        ...     start_inclusive=True,
        ...     end_inclusive=True,
        ...     has={"_ANY_EVENT": "(5, None)"},
        ...     index_timestamp="end",
        ... )
        >>> input_window.referenced_event
        ('trigger',)
        >>> # This window does not reference any "true" external predicates, only implicit predicates like
        >>> # start, end, and * events, so this list should be empty.
        >>> sorted(input_window.referenced_predicates)
        ['_ANY_EVENT']
        >>> input_window.start_endpoint_expr # doctest: +NORMALIZE_WHITESPACE
        ToEventWindowBounds(left_inclusive=True,
                            end_event='-_RECORD_START',
                            right_inclusive=True,
                            offset=datetime.timedelta(0))
        >>> input_window.end_endpoint_expr # doctest: +NORMALIZE_WHITESPACE
        TemporalWindowBounds(left_inclusive=False,
                             window_size=datetime.timedelta(days=2),
                             right_inclusive=False,
                             offset=datetime.timedelta(0))
        >>> input_window.root_node
        'end'
        >>> gap_window = WindowConfig(
        ...     start="input.end",
        ...     end="start + 24h",
        ...     start_inclusive=False,
        ...     end_inclusive=True,
        ...     has={"discharge": "(None, 0)", "death": "(None, 0)"}
        ... )
        >>> gap_window.referenced_event
        ('input', 'end')
        >>> sorted(gap_window.referenced_predicates)
        ['death', 'discharge']
        >>> gap_window.start_endpoint_expr is None
        True
        >>> gap_window.end_endpoint_expr # doctest: +NORMALIZE_WHITESPACE
        TemporalWindowBounds(left_inclusive=False,
                             window_size=datetime.timedelta(days=1),
                             right_inclusive=True,
                             offset=datetime.timedelta(0))
        >>> gap_window.root_node
        'start'
        >>> gap_window = WindowConfig(
        ...     start="input.end",
        ...     end="start + 0h",
        ...     start_inclusive=False,
        ...     end_inclusive=True,
        ...     has={"discharge": "(None, 0)", "death": "(None, 0)"}
        ... )
        >>> gap_window.referenced_event
        ('input', 'end')
        >>> sorted(gap_window.referenced_predicates)
        ['death', 'discharge']
        >>> gap_window.start_endpoint_expr is None
        True
        >>> gap_window.end_endpoint_expr is None # doctest: +NORMALIZE_WHITESPACE
        True
        >>> gap_window.root_node
        'start'
        >>> target_window = WindowConfig(
        ...     start="gap.end",
        ...     end="start -> discharge_or_death",
        ...     start_inclusive=False,
        ...     end_inclusive=True,
        ...     has={}
        ... )
        >>> target_window.referenced_event
        ('gap', 'end')
        >>> sorted(target_window.referenced_predicates)
        ['discharge_or_death']
        >>> target_window.start_endpoint_expr is None
        True
        >>> target_window.end_endpoint_expr # doctest: +NORMALIZE_WHITESPACE
        ToEventWindowBounds(left_inclusive=False,
                            end_event='discharge_or_death',
                            right_inclusive=True,
                            offset=datetime.timedelta(0))
        >>> target_window.root_node
        'start'
        >>> target_window = WindowConfig(
        ...     start="end",
        ...     end="gap.end <- discharge_or_death",
        ...     start_inclusive=False,
        ...     end_inclusive=True,
        ...     has={}
        ... )
        >>> target_window.referenced_event
        ('gap', 'end')
        >>> sorted(target_window.referenced_predicates)
        ['discharge_or_death']
        >>> target_window.start_endpoint_expr is None
        True
        >>> target_window.end_endpoint_expr # doctest: +NORMALIZE_WHITESPACE
        ToEventWindowBounds(left_inclusive=False,
                            end_event='-discharge_or_death',
                            right_inclusive=False,
                            offset=datetime.timedelta(0))
        >>> target_window.root_node
        'end'
        >>> invalid_window = WindowConfig(
        ...     start="gap.end gap.start",
        ...     end="start -> discharge_or_death",
        ...     start_inclusive=False,
        ...     end_inclusive=True,
        ...     has={}
        ... ) # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        ValueError: Window boundary reference must be either a valid alphanumeric/'_' string or a reference to
            another window's start or end event, formatted as a valid alphanumeric/'_' string, followed by
            '.start' or '.end'.
            Got: 'gap.end gap.start'
        >>> invalid_window = WindowConfig(
        ...     start="input",
        ...     end="start window -> discharge_or_death",
        ...     start_inclusive=False,
        ...     end_inclusive=True,
        ...     has={"discharge": "(None, 0)", "death": "(None, 0)"}
        ... ) # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        ValueError: Window boundary reference must be either a valid alphanumeric/'_' string or a reference
        to another window's start or end event, formatted as a valid alphanumeric/'_' string, followed by
        '.start' or '.end'. Got: 'start window'
        >>> invalid_window = WindowConfig(
        ...     start="input",
        ...     end="window.foo -> discharge_or_death",
        ...     start_inclusive=False,
        ...     end_inclusive=True,
        ...     has={"discharge": "(None, 0)", "death": "(None, 0)"}
        ... ) # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        ValueError: Window boundary reference must be either a valid alphanumeric/'_' string or a reference
        to another window's start or end event, formatted as a valid alphanumeric/'_' string, followed by
        '.start' or '.end'. Got: 'window.foo'
        >>> invalid_window = WindowConfig(
        ...     start=None, end=None, start_inclusive=True, end_inclusive=True, has={}
        ... )
        Traceback (most recent call last):
            ...
        ValueError: Window cannot progress from the start of the record to the end of the record.
        >>> invalid_window = WindowConfig(
        ...     start="input.end",
        ...     end="start - 2d",
        ...     start_inclusive=False,
        ...     end_inclusive=True,
        ...     has={"discharge": "(None, 0)", "death": "(None, 0)"}
        ... )
        Traceback (most recent call last):
            ...
        ValueError: Window start will not occur before window end! Got: input.end -> start - 2d
        >>> invalid_window = WindowConfig(
        ...     start="end -> predicate",
        ...     end="input.end",
        ...     start_inclusive=False,
        ...     end_inclusive=True,
        ...     has={"discharge": "(None, 0)", "death": "(None, 0)"}
        ... )
        Traceback (most recent call last):
            ...
        ValueError: Window start will not occur before window end! Got: end -> predicate -> input.end
        >>> invalid_window = WindowConfig(
        ...     start="end - 24h", end="start + 1d", start_inclusive=True, end_inclusive=True, has={}
        ... ) # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        ValueError: Exactly one of the start or end of the window must reference the other.
        Got: end - 24h -> start + 1d
        >>> invalid_window = WindowConfig(
        ...     start="input.end",
        ...     end="input.end + 2d",
        ...     start_inclusive=False,
        ...     end_inclusive=True,
        ...     has={"discharge": "(None, 0)", "death": "(None, 0)"}
        ... ) # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        ValueError: Exactly one of the start or end of the window must reference the other.
        Got: input.end -> input.end + 2d
        >>> invalid_window = WindowConfig(
        ...     start="input.end",
        ...     end="start + -24h",
        ...     start_inclusive=False,
        ...     end_inclusive=True,
        ...     has={"discharge": "(None, 0)", "death": "(None, 0)"}
        ... )
        Traceback (most recent call last):
            ...
        ValueError: Window boundary cannot contain both '+' and '-' operators.
        >>> invalid_window = WindowConfig(
        ...     start="input.end",
        ...     end="start + invalid time string.",
        ...     start_inclusive=False,
        ...     end_inclusive=True,
        ...     has={"discharge": "(None, 0)", "death": "(None, 0)"}
        ... )
        Traceback (most recent call last):
            ...
        ValueError: Failed to parse timedelta from window offset for 'invalid time string.'
        >>> target_window = WindowConfig(
        ...     start="gap.end",
        ...     end="start <-> discharge_or_death",
        ...     start_inclusive=False,
        ...     end_inclusive=True,
        ...     has={}
        ... )
        Traceback (most recent call last):
            ...
        ValueError: Window boundary cannot contain both '->' and '<-' operators.
        >>> invalid_window = WindowConfig(
        ...     start="input.end",
        ...     end="input.end + 2d",
        ...     start_inclusive=False,
        ...     end_inclusive=True,
        ...     has={"discharge": "(0)", "death": "(None, 0)"}
        ... ) # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        ValueError: Invalid constraint format: discharge.
        Expected format: '(min, max)'. Got: '(0)'
    """

    start: str | None
    end: str | None
    start_inclusive: bool
    end_inclusive: bool
    has: dict[str, str] = field(default_factory=dict)
    label: str | None = None
    index_timestamp: str | None = None

    @classmethod
    def _check_reference(cls, reference: str):
        """Checks to ensure referenced events are valid."""
        err_str = (
            "Window boundary reference must be either a valid alphanumeric/'_' string "
            "or a reference to another window's start or end event, formatted as a valid "
            f"alphanumeric/'_' string, followed by '.start' or '.end'. Got: '{reference}'"
        )

        if "." in reference:
            if reference.count(".") > 1:
                raise ValueError(err_str)
            window, event = reference.split(".")
            if event not in {"start", "end"} or not re.match(r"^\w+$", window):
                raise ValueError(err_str)
        elif not re.match(r"^\w+$", reference):
            raise ValueError(err_str)

    @classmethod
    def _parse_boundary(cls, boundary: str) -> dict[str, str]:
        if "->" in boundary or "<-" in boundary:
            if "->" in boundary and "<-" in boundary:
                raise ValueError("Window boundary cannot contain both '->' and '<-' operators.")
            elif "->" in boundary:
                ref, predicate = (x.strip() for x in boundary.split("->"))
            else:
                ref, predicate = (x.strip() for x in boundary.split("<-"))
                predicate = "-" + predicate

            cls._check_reference(ref)

            return {
                "referenced": ref,
                "offset": None,
                "event_bound": predicate,
                "occurs_before": "-" in predicate,
            }
        elif "+" in boundary or "-" in boundary:
            if "+" in boundary and "-" in boundary:
                raise ValueError("Window boundary cannot contain both '+' and '-' operators.")
            elif "+" in boundary:
                ref, offset = (x.strip() for x in boundary.split("+"))
            else:
                ref, offset = (x.strip() for x in boundary.split("-"))
                offset = "-" + offset

            cls._check_reference(ref)

            try:
                parsed_offset = parse_timedelta(offset)
                if parsed_offset == timedelta(0):
                    logger.warning(f"Window offset for {boundary} is zero; this may not be intended.")
                    return {"referenced": ref, "offset": None, "event_bound": None, "occurs_before": None}
            except (ValueError, TypeError) as e:
                raise ValueError(f"Failed to parse timedelta from window offset for '{offset}'") from e

            return {"referenced": ref, "offset": offset, "event_bound": None, "occurs_before": "-" in offset}
        else:
            ref = boundary.strip()
            cls._check_reference(ref)
            return {"referenced": ref, "offset": None, "event_bound": None, "occurs_before": None}

    def __post_init__(self):
        # Parse the has constraints from the string representation to the tuple representation
        if self.has is not None:
            for each_constraint in self.has:
                elements = self.has[each_constraint].strip("()").split(",")
                elements = [element.strip() for element in elements]
                if len(elements) != 2:
                    raise ValueError(
                        f"Invalid constraint format: {each_constraint}. "
                        f"Expected format: '(min, max)'. Got: '{self.has[each_constraint]}'"
                    )
                self.has[each_constraint] = tuple(
                    int(element) if element not in ("None", "") else None for element in elements
                )

        if self.start is None and self.end is None:
            raise ValueError("Window cannot progress from the start of the record to the end of the record.")

        if self.start is None:
            self._parsed_start = {
                "referenced": "end",
                "offset": None,
                "event_bound": f"-{START_OF_RECORD_KEY}",
                "occurs_before": True,
            }
        else:
            self._parsed_start = self._parse_boundary(self.start)

        if self.end is None:
            self._parsed_end = {
                "referenced": "start",
                "offset": None,
                "event_bound": END_OF_RECORD_KEY,
                "occurs_before": False,
            }
        else:
            self._parsed_end = self._parse_boundary(self.end)

        if self._parsed_start["referenced"] == "end" and self._parsed_end["referenced"] == "start":
            raise ValueError(
                "Exactly one of the start or end of the window must reference the other. "
                f"Got: {self.start} -> {self.end}"
            )
        elif self._parsed_start["referenced"] == "end":
            self._start_references_end = True
            # We use `is False` because it may be None, which is distinct from True or False
            if self._parsed_start["occurs_before"] is False:
                raise ValueError(
                    f"Window start will not occur before window end! Got: {self.start} -> {self.end}"
                )
        elif self._parsed_end["referenced"] == "start":
            self._start_references_end = False
            # We use `is True` because it may be None, which is distinct from True or False
            if self._parsed_end["occurs_before"] is True:
                raise ValueError(
                    f"Window start will not occur before window end! Got: {self.start} -> {self.end}"
                )
        else:
            raise ValueError(
                "Exactly one of the start or end of the window must reference the other. "
                f"Got: {self.start} -> {self.end}"
            )

    @property
    def root_node(self) -> str:
        """Returns 'start' if the end of the window is defined relative to the start and 'end' otherwise."""
        return "end" if self._start_references_end else "start"

    @property
    def referenced_event(self) -> tuple[str]:
        if self._start_references_end:
            return tuple(self._parsed_end["referenced"].split("."))
        else:
            return tuple(self._parsed_start["referenced"].split("."))

    @property
    def constraint_predicates(self) -> set[str]:
        predicates = set(self.has.keys())
        return predicates

    @property
    def referenced_predicates(self) -> set[str]:
        predicates = set(self.has.keys())
        if self._parsed_start["event_bound"]:
            predicates.add(self._parsed_start["event_bound"].replace("-", ""))
        if self._parsed_end["event_bound"]:
            predicates.add(self._parsed_end["event_bound"].replace("-", ""))

        predicates -= {START_OF_RECORD_KEY, END_OF_RECORD_KEY}
        return predicates

    @property
    def start_endpoint_expr(self) -> None | ToEventWindowBounds | TemporalWindowBounds:
        if self._start_references_end:
            # If end references start, then end will occur after start, so `left_inclusive` corresponds to
            # `start_inclusive` and `right_inclusive` corresponds to `end_inclusive`.
            left_inclusive = self.start_inclusive
            right_inclusive = self.end_inclusive
        else:
            # If this window references end from start, then the end event window expression will not have
            # any constraints as it will reference an external event, and therefore the inclusive
            # parameters don't matter.
            left_inclusive = False
            right_inclusive = False

        if self._parsed_start["event_bound"]:
            return ToEventWindowBounds(
                end_event=self._parsed_start["event_bound"],
                left_inclusive=left_inclusive,
                right_inclusive=right_inclusive,
            )
        elif self._parsed_start["offset"]:
            return TemporalWindowBounds(
                window_size=parse_timedelta(self._parsed_start["offset"]),
                left_inclusive=left_inclusive,
                right_inclusive=right_inclusive,
            )
        else:
            return None

    @property
    def end_endpoint_expr(self) -> None | ToEventWindowBounds | TemporalWindowBounds:
        if self._start_references_end:
            # If this window references end from start, then the end event window expression will not have
            # any constraints as it will reference an external event, and therefore the inclusive
            # parameters don't matter.
            left_inclusive = False
            right_inclusive = False
        else:
            # If end references start, then end will occur after start, so `left_inclusive` corresponds to
            # `start_inclusive` and `right_inclusive` corresponds to `end_inclusive`.
            left_inclusive = self.start_inclusive
            right_inclusive = self.end_inclusive

        if self._parsed_end["event_bound"]:
            return ToEventWindowBounds(
                end_event=self._parsed_end["event_bound"],
                left_inclusive=left_inclusive,
                right_inclusive=right_inclusive,
            )
        elif self._parsed_end["offset"]:
            return TemporalWindowBounds(
                window_size=parse_timedelta(self._parsed_end["offset"]),
                left_inclusive=left_inclusive,
                right_inclusive=right_inclusive,
            )
        else:
            return None


@dataclasses.dataclass
class EventConfig:
    """A configuration object for defining the event that triggers the task extraction process.

    This is defined by all events that match a simple predicate. This event serves as the root of the window
    tree, and its form is dictated by the fact that we must be able to localize the tree to identify valid
    realizations of the tree.

    Examples:
        >>> event = EventConfig("event_type//ADMISSION")
        >>> event.predicate
        'event_type//ADMISSION'
    """

    predicate: str


@dataclasses.dataclass
class TaskExtractorConfig:
    """A configuration object for parsing the plain-data stored in a task extractor config.

    This class can be serialized to and deserialized from a YAML file, and is largely a collection of
    utilities to parse, validate, and leverage task extraction configuration data in practice. There is no
    state stored in this class that is not present or recoverable from the source YAML file on disk. It also
    can be read from a simplified, "user-friendly" language, which can also be stored on or read from disk,
    which is ultimately parsed into the expansive, full specification contained in the YAML file referenced
    above.

    Args:
        predicates: A dictionary of predicate configurations, stored as either plain or derived predicate
            configuration objects (which are simple dataclasses with utility functions over plain
            dictionaries).
        trigger: The event configuration that triggers the task extraction process. This is a simple
            dataclass with a single field, the name of the predicate that triggers the task extraction and
            serves as the root of the window tree.
        windows: A dictionary of window configurations. Each window configuration is a simple dataclass with
            that can be materialized to/from a simple, POD dictionary.

    Raises:
        ValueError: If any window or predicate names are not composed of alphanumeric or "_" characters.

    Examples:
        >>> from bigtree import print_tree
        >>> predicates = {
        ...     "admission": PlainPredicateConfig("admission"),
        ...     "discharge": PlainPredicateConfig("discharge"),
        ...     "death": PlainPredicateConfig("death"),
        ...     "death_or_discharge": DerivedPredicateConfig("or(death, discharge)"),
        ...     "diabetes_icd9": PlainPredicateConfig("ICD9CM//250.02"),
        ...     "diabetes_icd10": PlainPredicateConfig("ICD10CM//E11.65"),
        ...     "diabetes": DerivedPredicateConfig("or(diabetes_icd9, diabetes_icd10)"),
        ...     "diabetes_and_discharge": DerivedPredicateConfig("and(diabetes, discharge)"),
        ... }
        >>> trigger = EventConfig("admission")
        >>> windows = {
        ...     "input": WindowConfig(
        ...         start=None,
        ...         end="trigger + 24h",
        ...         start_inclusive=True,
        ...         end_inclusive=True,
        ...         has={"_ANY_EVENT": "(32, None)"},
        ...         index_timestamp="end",
        ...     ),
        ...     "gap": WindowConfig(
        ...         start="input.end",
        ...         end="start + 24h",
        ...         start_inclusive=False,
        ...         end_inclusive=True,
        ...         has={"death_or_discharge": "(None, 0)", "admission": "(None, 0)"},
        ...     ),
        ...     "target": WindowConfig(
        ...         start="gap.end",
        ...         end="start -> death_or_discharge",
        ...         start_inclusive=False,
        ...         end_inclusive=True,
        ...         has={},
        ...         label="death",
        ...     ),
        ... }
        >>> config = TaskExtractorConfig(predicates=predicates, trigger=trigger, windows=windows)
        >>> print(config.plain_predicates) # doctest: +NORMALIZE_WHITESPACE
        {'admission': PlainPredicateConfig(code='admission',
                              value_min=None,
                              value_max=None,
                              value_min_inclusive=None,
                              value_max_inclusive=None,
                              static=False,
                              other_cols={}),
         'discharge': PlainPredicateConfig(code='discharge',
                              value_min=None,
                              value_max=None,
                              value_min_inclusive=None,
                              value_max_inclusive=None,
                              static=False,
                              other_cols={}),
         'death': PlainPredicateConfig(code='death',
                              value_min=None,
                              value_max=None,
                              value_min_inclusive=None,
                              value_max_inclusive=None,
                              static=False,
                              other_cols={}),
         'diabetes_icd9': PlainPredicateConfig(code='ICD9CM//250.02',
                              value_min=None,
                              value_max=None,
                              value_min_inclusive=None,
                              value_max_inclusive=None,
                              static=False,
                              other_cols={}),
         'diabetes_icd10': PlainPredicateConfig(code='ICD10CM//E11.65',
                              value_min=None,
                              value_max=None,
                              value_min_inclusive=None,
                              value_max_inclusive=None,
                              static=False,
                              other_cols={})}
        >>> print(config.label_window) # doctest: +NORMALIZE_WHITESPACE
        target
        >>> print(config.index_timestamp_window) # doctest: +NORMALIZE_WHITESPACE
        input
        >>> print(config.derived_predicates) # doctest: +NORMALIZE_WHITESPACE
        {'death_or_discharge': DerivedPredicateConfig(expr='or(death, discharge)', static=False),
         'diabetes': DerivedPredicateConfig(expr='or(diabetes_icd9, diabetes_icd10)', static=False),
         'diabetes_and_discharge': DerivedPredicateConfig(expr='and(diabetes, discharge)', static=False)}
        >>> print(nx.write_network_text(config.predicates_DAG))
        ╟── death
        ╎   └─╼ death_or_discharge ╾ discharge
        ╟── discharge
        ╎   ├─╼ diabetes_and_discharge ╾ diabetes
        ╎   └─╼  ...
        ╟── diabetes_icd9
        ╎   └─╼ diabetes ╾ diabetes_icd10
        ╎       └─╼  ...
        ╙── diabetes_icd10
            └─╼  ...
        >>> print_tree(config.window_tree)
        trigger
        └── input.end
            ├── input.start
            └── gap.end
                └── target.end

    Configs will error out in various ways when passed inappropriate arguments:
        >>> config_path = "/foo/non_existent_file.yaml"
        >>> cfg = TaskExtractorConfig.load(config_path)
        Traceback (most recent call last):
            ...
        FileNotFoundError: Cannot load missing configuration file /foo/non_existent_file.yaml!

        >>> import tempfile
        >>> with tempfile.NamedTemporaryFile(mode="w", suffix=".txt") as f:
        ...     config_path = Path(f.name)
        ...     cfg = TaskExtractorConfig.load(config_path)
        Traceback (most recent call last):
            ...
        ValueError: Only supports reading from '.yaml'. Got: '.txt' in ....txt'.
        >>> predicates_path = "/foo/non_existent_predicates.yaml"
        >>> with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml") as f:
        ...     config_path = Path(f.name)
        ...     cfg = TaskExtractorConfig.load(config_path, predicates_path)
        Traceback (most recent call last):
            ...
        FileNotFoundError: Cannot load missing predicates file /foo/non_existent_predicates.yaml!
        >>> with tempfile.NamedTemporaryFile(mode="w", suffix=".txt") as f:
        ...     predicates_path = Path(f.name)
        ...     with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml") as f2:
        ...         config_path = Path(f2.name)
        ...         cfg = TaskExtractorConfig.load(config_path, predicates_path)
        Traceback (most recent call last):
            ...
        ValueError: Only supports reading from '.yaml'. Got: '.txt' in ....txt'.
        >>> import yaml
        >>> data = {
        ...     'predicates': {},
        ...     'trigger': {},
        ...     'foo': {}
        ... }
        >>> with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml") as f:
        ...     config_path = Path(f.name)
        ...     yaml.dump(data, f)
        ...     cfg = TaskExtractorConfig.load(config_path)
        Traceback (most recent call last):
            ...
        ValueError: Unrecognized keys in configuration file: 'foo'
        >>> with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml") as f:
        ...     predicates_path = Path(f.name)
        ...     yaml.dump(data, f)
        ...     with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml") as f2:
        ...         config_path = Path(f2.name)
        ...         cfg = TaskExtractorConfig.load(config_path, predicates_path)
        Traceback (most recent call last):
            ...
        ValueError: Unrecognized keys in configuration file: 'foo, trigger'

        >>> predicates = {"foo bar": PlainPredicateConfig("foo")}
        >>> trigger = EventConfig("foo")
        >>> config = TaskExtractorConfig(predicates=predicates, trigger=trigger, windows={})
        Traceback (most recent call last):
            ...
        ValueError: Predicate name 'foo bar' is invalid; must be composed of alphanumeric or '_' characters.
        >>> predicates = {"foo": str("foo")}
        >>> trigger = EventConfig("foo")
        >>> config = TaskExtractorConfig(predicates=predicates, trigger=trigger, windows={})
        ... # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        ValueError: Invalid predicate configuration for 'foo': foo. Must be either a PlainPredicateConfig or
        DerivedPredicateConfig object. Got: <class 'str'>
        >>> predicates = {
        ...     "foo": PlainPredicateConfig("foo"),
        ...     "foobar": DerivedPredicateConfig("or(foo, bar)"),
        ... }
        >>> trigger = EventConfig("foo")
        >>> config = TaskExtractorConfig(predicates=predicates, trigger=trigger, windows={})
        Traceback (most recent call last):
            ...
        KeyError: "Missing 1 relationships:\\nDerived predicate 'foobar' references undefined predicate 'bar'"

        >>> predicates = {"foo": PlainPredicateConfig("foo")}
        >>> trigger = EventConfig("foo")
        >>> windows = {"foo bar": WindowConfig("gap.end", "start + 24h", True, True)}
        >>> config = TaskExtractorConfig(predicates=predicates, trigger=trigger, windows=windows)
        Traceback (most recent call last):
            ...
        ValueError: Window name 'foo bar' is invalid; must be composed of alphanumeric or '_' characters.
        >>> windows = {"foo": WindowConfig("gap.end", "start + 24h", True, True, {}, "bar")}
        >>> config = TaskExtractorConfig(predicates=predicates, trigger=trigger, windows=windows)
        Traceback (most recent call last):
            ...
        ValueError: Label must be one of the defined predicates. Got: bar for window 'foo'
        >>> windows = {"foo": WindowConfig("gap.end", "start + 24h", True, True, {}, "foo", "bar")}
        >>> config = TaskExtractorConfig(predicates=predicates, trigger=trigger, windows=windows)
        Traceback (most recent call last):
            ...
        ValueError: Index timestamp must be either 'start' or 'end'. Got: bar for window 'foo'
        >>> windows = {
        ...     "foo": WindowConfig("gap.end", "start + 24h", True, True, {}, "foo"),
        ...     "bar": WindowConfig("gap.end", "start + 24h", True, True, {}, "foo")
        ... }
        >>> config = TaskExtractorConfig(predicates=predicates, trigger=trigger, windows=windows)
        Traceback (most recent call last):
            ...
        ValueError: Only one window can be labeled, found 2 labeled windows: foo, bar
        >>> windows = {
        ...     "foo": WindowConfig("gap.end", "start + 24h", True, True, {}, "foo", "start"),
        ...     "bar": WindowConfig("gap.end", "start + 24h", True, True, {}, index_timestamp="start")
        ... }
        >>> config = TaskExtractorConfig(predicates=predicates, trigger=trigger, windows=windows)
        ... # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        ValueError: Only the 'start'/'end' of one window can be used as the index timestamp, found
        2 windows with index_timestamp: foo, bar

        >>> predicates = {"foo": PlainPredicateConfig("foo")}
        >>> trigger = EventConfig("bar")
        >>> config = TaskExtractorConfig(predicates=predicates, trigger=trigger, windows={})
        Traceback (most recent call last):
            ...
        KeyError: "Trigger event predicate 'bar' not found in predicates: foo"
    """

    predicates: dict[str, PlainPredicateConfig | DerivedPredicateConfig]
    trigger: EventConfig
    windows: dict[str, WindowConfig] | None
    label_window: str | None = None
    index_timestamp_window: str | None = None

    @classmethod
    def load(cls, config_path: str | Path, predicates_path: str | Path = None) -> TaskExtractorConfig:
        """Load a configuration file from the given path and return it as a dict.

        Args:
            config_path: The path to which a configuration object will be read from in YAML form.

        Raises:
            FileNotFoundError: If the file does not exist.
            ValueError: If the file is not a ".yaml" file.

        Examples:
            >>> import tempfile
            >>> yaml = ruamel.yaml.YAML(typ="safe", pure=True)
            >>> config_dict = {
            ...     "metadata": {'description': 'A test configuration file'},
            ...     "description": 'this is a test',
            ...     "predicates": {"admission": {"code": "admission"}},
            ...     "trigger": "admission",
            ...     "windows": {
            ...         "start": {
            ...             "start": None, "end": "trigger + 24h", "start_inclusive": True,
            ...             "end_inclusive": True,
            ...         }
            ...     },
            ... }
            >>> with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml") as f:
            ...     config_path = Path(f.name)
            ...     yaml.dump(config_dict, f)
            ...     cfg = TaskExtractorConfig.load(config_path)
            >>> cfg # doctest: +NORMALIZE_WHITESPACE
            TaskExtractorConfig(predicates={'admission': PlainPredicateConfig(code='admission',
                                                           value_min=None, value_max=None,
                                                           value_min_inclusive=None, value_max_inclusive=None,
                                                           static=False, other_cols={})},
                                trigger=EventConfig(predicate='admission'),
                                windows={'start': WindowConfig(start=None, end='trigger + 24h',
                                                    start_inclusive=True, end_inclusive=True, has={},
                                                    label=None, index_timestamp=None)},
                                label_window=None, index_timestamp_window=None)
            >>> predicates_dict = {
            ...     "metadata": {'description': 'A test predicates file'},
            ...     "description": 'this is a test',
            ...     "patient_demographics": {"brown_eyes": {"code": "eye_color//BR"}},
            ...     "predicates": {"admission": {"code": "admission"}},
            ... }
            >>> no_predicates_config = {k: v for k, v in config_dict.items() if k != "predicates"}
            >>> with (tempfile.NamedTemporaryFile(mode="w", suffix=".yaml") as config_fp,
            ...      tempfile.NamedTemporaryFile(mode="w", suffix=".yaml") as pred_fp):
            ...     config_path = Path(config_fp.name)
            ...     pred_path = Path(pred_fp.name)
            ...     yaml.dump(no_predicates_config, config_fp)
            ...     yaml.dump(predicates_dict, pred_fp)
            ...     cfg = TaskExtractorConfig.load(config_path, pred_path)
            >>> cfg # doctest: +NORMALIZE_WHITESPACE
            TaskExtractorConfig(predicates={'admission': PlainPredicateConfig(code='admission',
                                                           value_min=None, value_max=None,
                                                           value_min_inclusive=None, value_max_inclusive=None,
                                                           static=False, other_cols={}),
                                            'brown_eyes': PlainPredicateConfig(code='eye_color//BR',
                                                            value_min=None, value_max=None,
                                                            value_min_inclusive=None,
                                                            value_max_inclusive=None, static=True,
                                                            other_cols={})},
                                trigger=EventConfig(predicate='admission'),
                                windows={'start': WindowConfig(start=None, end='trigger + 24h',
                                                    start_inclusive=True, end_inclusive=True, has={},
                                                    label=None, index_timestamp=None)},
                                label_window=None, index_timestamp_window=None)
        """
        if isinstance(config_path, str):
            config_path = Path(config_path)

        if not config_path.is_file():
            raise FileNotFoundError(f"Cannot load missing configuration file {str(config_path.resolve())}!")

        if config_path.suffix == ".yaml":
            yaml = ruamel.yaml.YAML(typ="safe", pure=True)
            loaded_dict = yaml.load(config_path.read_text())
        else:
            raise ValueError(
                f"Only supports reading from '.yaml'. Got: '{config_path.suffix}' in '{config_path.name}'."
            )

        overriding_predicates = {}
        overriding_demographics = {}
        if predicates_path:
            if isinstance(predicates_path, str):
                predicates_path = Path(predicates_path)

            if not predicates_path.is_file():
                raise FileNotFoundError(
                    f"Cannot load missing predicates file {str(predicates_path.resolve())}!"
                )

            if predicates_path.suffix == ".yaml":
                yaml = ruamel.yaml.YAML(typ="safe", pure=True)
                predicates_dict = yaml.load(predicates_path.read_text())
            else:
                raise ValueError(
                    f"Only supports reading from '.yaml'. Got: '{predicates_path.suffix}' in "
                    f"'{predicates_path.name}'."
                )

            # Remove the description or metadata keys if they exist - currently unused except for readability
            # in the YAML
            _ = predicates_dict.pop("description", None)
            _ = predicates_dict.pop("metadata", None)
            overriding_predicates = predicates_dict.pop("predicates", {})
            overriding_demographics = predicates_dict.pop("patient_demographics", {})

            if predicates_dict:
                raise ValueError(
                    f"Unrecognized keys in configuration file: '{', '.join(predicates_dict.keys())}'"
                )

        # Remove the description or metadata keys if they exist - currently unused except for readability
        # in the YAML
        _ = loaded_dict.pop("description", None)
        _ = loaded_dict.pop("metadata", None)

        trigger = loaded_dict.pop("trigger")
        windows = loaded_dict.pop("windows", None)

        predicates = loaded_dict.pop("predicates", {})
        patient_demographics = loaded_dict.pop("patient_demographics", {})

        if loaded_dict:
            raise ValueError(f"Unrecognized keys in configuration file: '{', '.join(loaded_dict.keys())}'")

        final_predicates = {**predicates, **overriding_predicates}
        final_demographics = {**patient_demographics, **overriding_demographics}

        logger.info("Parsing windows...")
        if windows is None:
            windows = {}
            logger.warning(
                "No windows specified in configuration file. Extracting only matching trigger events."
            )
        else:
            windows = {n: WindowConfig(**w) for n, w in windows.items()}

        logger.info("Parsing trigger event...")
        trigger = EventConfig(trigger)

        referenced_predicates = {pred for w in windows.values() for pred in w.referenced_predicates}
        referenced_predicates.add(trigger.predicate)
        current_predicates = set(referenced_predicates)
        special_predicates = {ANY_EVENT_COLUMN, START_OF_RECORD_KEY, END_OF_RECORD_KEY}
        for pred in current_predicates - special_predicates:
            if pred not in final_predicates:
                raise KeyError(
                    f"Something referenced predicate {pred} that wasn't defined in the configuration."
                )
            if "expr" in final_predicates[pred]:
                referenced_predicates.update(
                    DerivedPredicateConfig(**final_predicates[pred]).input_predicates
                )

        logger.info("Parsing predicates...")
        predicates_to_parse = {k: v for k, v in final_predicates.items() if k in referenced_predicates}
        predicate_objs = {}
        for n, p in predicates_to_parse.items():
            if "expr" in p:
                predicate_objs[n] = DerivedPredicateConfig(**p)
            else:
                config_data = {k: v for k, v in p.items() if k in PlainPredicateConfig.__dataclass_fields__}
                other_cols = {k: v for k, v in p.items() if k not in config_data}
                predicate_objs[n] = PlainPredicateConfig(**config_data, other_cols=other_cols)

        if final_demographics:
            logger.info("Parsing patient demographics...")
            final_demographics = {
                n: PlainPredicateConfig(**p, static=True) for n, p in final_demographics.items()
            }
            predicate_objs.update(final_demographics)

        return cls(predicates=predicate_objs, trigger=trigger, windows=windows)

    def _initialize_predicates(self):
        """Initialize the predicates tree from the configuration object and check validity.

        Raises:
            ValueError: If the predicate name is not valid.
        """

        dag_relationships = []

        for name, predicate in self.predicates.items():
            if re.match(r"^\w+$", name) is None:
                raise ValueError(
                    f"Predicate name '{name}' is invalid; must be composed of alphanumeric or '_' characters."
                )

            match predicate:
                case PlainPredicateConfig():
                    pass
                case DerivedPredicateConfig():
                    for pred in predicate.input_predicates:
                        dag_relationships.append((pred, name))
                case _:
                    raise ValueError(
                        f"Invalid predicate configuration for '{name}': {predicate}. "
                        "Must be either a PlainPredicateConfig or DerivedPredicateConfig object. "
                        f"Got: {type(predicate)}"
                    )

        missing_predicates = []
        for parent, child in dag_relationships:
            if parent not in self.predicates:
                missing_predicates.append(
                    f"Derived predicate '{child}' references undefined predicate '{parent}'"
                )
        if missing_predicates:
            raise KeyError(
                f"Missing {len(missing_predicates)} relationships:\n" + "\n".join(missing_predicates)
            )

        self._predicate_dag_graph = nx.DiGraph(dag_relationships)
        if not nx.is_directed_acyclic_graph(self._predicate_dag_graph):
            raise ValueError(
                "Predicate graph is not a directed acyclic graph!\n"
                f"Cycle found: {nx.find_cycle(self._predicate_dag_graph)}\n"
                f"Graph: {nx.write_network_text(self._predicate_dag_graph)}"
            )

    def _initialize_windows(self):
        """Initialize the windows tree from the configuration object and check validity.

        Raises:
            ValueError: If the window name is not valid.
        """

        for name in self.windows:
            if re.match(r"^\w+$", name) is None:
                raise ValueError(
                    f"Window name '{name}' is invalid; must be composed of alphanumeric or '_' characters."
                )

        label_windows = []
        index_timestamp_windows = []
        for name, window in self.windows.items():
            if window.label:
                if window.label not in self.predicates:
                    raise ValueError(
                        f"Label must be one of the defined predicates. Got: {window.label} "
                        f"for window '{name}'"
                    )
                label_windows.append(name)
            if window.index_timestamp:
                if window.index_timestamp not in ["start", "end"]:
                    raise ValueError(
                        f"Index timestamp must be either 'start' or 'end'. Got: {window.index_timestamp} "
                        f"for window '{name}'"
                    )
                index_timestamp_windows.append(name)
        if len(label_windows) > 1:
            raise ValueError(
                f"Only one window can be labeled, found {len(label_windows)} labeled windows: "
                f"{', '.join(label_windows)}"
            )
        self.label_window = label_windows[0] if label_windows else None
        if len(index_timestamp_windows) > 1:
            raise ValueError(
                f"Only the 'start'/'end' of one window can be used as the index timestamp, "
                f"found {len(index_timestamp_windows)} windows with index_timestamp: "
                f"{', '.join(index_timestamp_windows)}"
            )
        self.index_timestamp_window = index_timestamp_windows[0] if index_timestamp_windows else None

        if self.trigger.predicate not in self.predicates and self.trigger.predicate not in [
            ANY_EVENT_COLUMN,
            START_OF_RECORD_KEY,
            END_OF_RECORD_KEY,
        ]:
            raise KeyError(
                f"Trigger event predicate '{self.trigger.predicate}' not found in predicates: "
                f"{', '.join(self.predicates.keys())}"
            )

        trigger_node = Node("trigger")

        window_nodes = {"trigger": trigger_node}
        for name, window in self.windows.items():
            start_node = Node(f"{name}.start", endpoint_expr=window.start_endpoint_expr)
            end_node = Node(f"{name}.end", endpoint_expr=window.end_endpoint_expr)

            if window.root_node == "end":
                # In this case, the end_node will bound an unconstrained window, as it is the window between
                # a prior window and the defined anchor for this window, so it has no constraints. But the
                # start_node will have the constraints corresponding to this window, as it is defined relative
                # to the end node.
                end_node.constraints = {}
                start_node.constraints = window.has
                start_node.parent = end_node
            else:
                # In this case, the start_node will bound an unconstrained window, as it is the window between
                # a prior window and the defined anchor for this window, so it has no constraints. But the
                # start_node will have the constraints corresponding to this window, as it is defined relative
                # to the end node.
                end_node.constraints = window.has
                start_node.constraints = {}
                end_node.parent = start_node

            window_nodes[f"{name}.start"] = start_node
            window_nodes[f"{name}.end"] = end_node

        for name, window in self.windows.items():
            for predicate in window.referenced_predicates - {ANY_EVENT_COLUMN}:
                if predicate not in self.predicates:
                    raise KeyError(
                        f"Window '{name}' references undefined predicate '{predicate}'.\n"
                        f"Window predicates: {', '.join(window.referenced_predicates)}\n"
                        f"Defined predicates: {', '.join(self.predicates.keys())}"
                    )

            if len(window.referenced_event) == 1:
                event = window.referenced_event[0]
                if event != "trigger":
                    raise KeyError(
                        f"Window '{name}' references undefined trigger event '{event}' -- must be trigger!"
                    )

                window_nodes[f"{name}.{window.root_node}"].parent = window_nodes[event]

            elif len(window.referenced_event) == 2:
                referenced_window, referenced_event = window.referenced_event
                if referenced_window not in self.windows:
                    raise KeyError(
                        f"Window '{name}' references undefined window '{referenced_window}' "
                        f"for event '{referenced_event}'. Allowed windows: {', '.join(self.windows.keys())}"
                    )
                if referenced_event not in {"start", "end"}:
                    raise KeyError(
                        f"Window '{name}' references undefined event '{referenced_event}' "
                        f"for window '{referenced_window}'. Allowed events: 'start', 'end'"
                    )

                parent_node = f"{referenced_window}.{referenced_event}"
                window_nodes[f"{name}.{window.root_node}"].parent = window_nodes[parent_node]
            else:
                raise ValueError(
                    f"Window '{name}' references invalid event '{window.referenced_event}' "
                    "must be of length 1 or 2."
                )

        # Clean up the tree
        nodes_to_remove = []

        # First pass: identify nodes to remove, reassign children's parent
        for n, node in window_nodes.items():
            if n != "trigger" and node.endpoint_expr is None:
                nodes_to_remove.append(n)
                for child in node.children:
                    # Reassign
                    child.parent = node.parent
                    if node.parent:
                        if child not in node.parent.children:
                            node.parent.children += (child,)

        # Second pass: remove nodes from parent's children
        for node_name in nodes_to_remove:
            node = window_nodes[node_name]
            if node.parent:
                # Remove
                node.parent.children = [child for child in node.parent.children if child.name != node_name]

        # Delete nodes_to_remove
        for node_name in nodes_to_remove:
            del window_nodes[node_name]

        self.window_nodes = window_nodes

    def __post_init__(self):
        self._initialize_predicates()
        self._initialize_windows()

    @property
    def window_tree(self) -> Node:
        return self.window_nodes["trigger"]

    @property
    def predicates_DAG(self) -> nx.DiGraph:
        return self._predicate_dag_graph

    @property
    def plain_predicates(self) -> dict[str:PlainPredicateConfig]:
        """Returns a dictionary of plain predicates in {name: code} format."""
        return {p: cfg for p, cfg in self.predicates.items() if cfg.is_plain}

    @property
    def derived_predicates(self) -> OrderedDict[str:DerivedPredicateConfig]:
        """Returns an ordered dictionary mapping derived predicates to their configs in a proper order."""
        return {
            p: self.predicates[p]
            for p in nx.topological_sort(self.predicates_DAG)
            if not self.predicates[p].is_plain
        }



"""Tests the full end-to-end extraction process."""


import rootutils

root = rootutils.setup_root(__file__, dotenv=True, pythonpath=True, cwd=True)


import polars as pl

from .test_meds import parse_labels_yaml, parse_shards_yaml
from .utils import cli_test

pl.enable_string_cache()

# Data (input)
MEDS_SHARDS = parse_shards_yaml(
    """
  "0": |-2
    subject_id,time,code,numeric_value,text_value
    1,,GENDER//MALE,,
    1,,SNP//rs234567,,
    1,12/18/1960 11:03,MEDS_BIRTH,,
    1,08/02/1972 10:00,CLINIC_VISIT,,
    1,08/02/1972 10:00,ICD9CM//493.90,,
    1,08/02/1972 10:00,LOINC//8310-5,0.65,
    1,08/02/1972 10:00,VITALS//BP//SYSTOLIC,108,
    1,01/14/2020 15:14,ADMISSION//MEDICAL,,
    1,01/14/2020 15:18,VITALS//BP//SYSTOLIC,132,
    1,01/14/2020 15:18,VITALS//BP//DIASTOLIC,90,
    1,01/14/2020 15:18,VITALS//HR//BPM,121,
    1,01/14/2020 15:18,VITALS//WEIGHT//LBS,233.2,
    1,01/15/2020 10:04,VITALS//BP//SYSTOLIC,126,
    1,01/15/2020 10:04,VITALS//BP//DIASTOLIC,91,
    1,01/15/2020 10:04,VITALS//HR//BPM,85,
    1,01/16/2020 10:11,VITALS//BP//SYSTOLIC,135,
    1,01/16/2020 10:11,VITALS//BP//DIASTOLIC,88,
    1,01/16/2020 10:11,VITALS//HR//BPM,79,
    1,01/16/2020 13:02,LVEF//ECHO,0.24,
    1,01/17/2020 10:00,ICD9CM//428.9,,
    1,01/17/2020 10:00,DISCHARGE//HOME,,
    1,01/18/2022 04:46,ADMISSION//MEDICAL,,
    1,01/20/2022 08:00,DISCHARGE//HOME_AMA,,
    1,01/20/2022 08:00,ICD9CM//428.41,,
    1,01/20/2022 08:00,ICD9CM//451.1,,
    1,01/24/2022 08:11,ADMISSION//ED,,
    1,01/25/2022 10:04,VITALS//BP//SYSTOLIC,168,
    1,01/25/2022 10:04,VITALS//BP//DIASTOLIC,100,
    1,01/25/2022 10:04,VITALS//HR//BPM,56,
    1,02/27/2022 01:13,ICD9CM//428.41,,
    1,02/27/2022 01:13,ICD9CM//410.1,,
    1,02/27/2022 01:13,DEATH,,

  "1": |-2
    subject_id,time,code,numeric_value,text_value
    3,,GENDER//FEMALE,,
    3,,SNP//rs2345291,,
    3,,SNP//rs228192,,
    3,02/28/1982 00:00,MEDS_BIRTH,,
    3,01/14/2020 15:14,ADMISSION//MEDICAL,,
    3,01/14/2020 15:18,VITALS//BP//SYSTOLIC,132,
    3,01/14/2020 15:18,VITALS//BP//DIASTOLIC,90,
    3,01/14/2020 15:18,VITALS//HR//BPM,121,
    3,01/17/2020 10:00,ICD9CM//V30.00,,
    3,01/17/2020 10:00,DISCHARGE//HOME,,
    3,01/18/2020 18:18,ADMISSION//MEDICAL,,
    3,01/20/2020 15:18,DISCHARGE//HOME,,
    3,03/18/2024 16:54,ICD9CM//428.9,,
    3,03/18/2024 17:11,ADMISSION//SURGICAL,,
    3,03/28/2024 10:00,DISCHARGE//HOME,,
    3,03/29/2024 11:00,ADMISSION//SURGICAL,,
    3,04/19/2024 13:32,DISCHARGE//HOME,,
    3,05/22/2024 00:00,ICD9CM//428.9,,
    """,
    text_value=pl.Utf8,
)

# Tasks (input)
TASKS = {
    "inhospital_mortality": """
        predicates:
          admission:
            code: {regex: ADMISSION//.*}
          discharge:
            code: {regex: DISCHARGE//.*}
          death:
            code: DEATH
          discharge_or_death:
            expr: or(discharge, death)

        trigger: admission

        windows:
          input:
            start: NULL
            end: trigger + 24h
            start_inclusive: True
            end_inclusive: True
            has:
              _ANY_EVENT: (5, None)
            index_timestamp: end
          gap:
            start: trigger
            end: start + 48h
            start_inclusive: False
            end_inclusive: True
            has:
              admission: (None, 0)
              discharge_or_death: (None, 0)
          target:
            start: gap.end
            end: start -> discharge_or_death
            start_inclusive: False
            end_inclusive: True
            label: death
        """,
    "HF_derived_readmission": """
        predicates:
          admission:
            code: {regex: ADMISSION//.*}
          discharge:
            code: {regex: DISCHARGE//.*}
          HF_dx:
            code: {regex: ICD9CM//428.*}

        trigger: discharge

        windows:
          data_within_5yr_of_admit:
            start: end - 1825d
            end: admission_is_HF.start
            start_inclusive: True
            end_inclusive: False
            has:
              _ANY_EVENT: (1, None)
          admission_is_HF:
            start: end <- admission
            end: trigger
            start_inclusive: True
            end_inclusive: True
            has:
              HF_dx: (1, None)
          input:
            start: NULL
            end: trigger
            start_inclusive: True
            end_inclusive: True
            index_timestamp: end
          target:
            start: input.end
            end: start + 30d
            start_inclusive: False
            end_inclusive: True
            label: admission
          censor_protection:
            start: target.end
            end: null
            start_inclusive: False
            end_inclusive: True
            has:
              _ANY_EVENT: (1, None)
    """,
    "nested_preds_readmission": """
        predicates:
          hospital_admission_0:
            code: ADMISSION//ED
          hospital_admission_1:
            code: ADMISSION//EU OBSERVATION//EMERGENCY ROOM
          hospital_admission_2:
            code: ADMISSION//SURGICAL
          hospital_admission_3:
            code: ADMISSION//OBSERVATION ADMIT//EMERGENCY ROOM
          hospital_admission_4:
            code: ADMISSION//URGENT//TRANSFER FROM HOSPITAL
          hospital_admission_5:
            code: ADMISSION//URGENT//PHYSICIAN REFERRAL
          hospital_admission_6:
            code: ADMISSION//DIRECT EMER.//PHYSICIAN REFERRAL
          hospital_admission_7:
            code: ADMISSION//OBSERVATION ADMIT//PHYSICIAN REFERRAL
          hospital_admission_8:
            code: ADMISSION//DIRECT OBSERVATION//PHYSICIAN REFERRAL
          hospital_admission_9:
            code: ADMISSION//ELECTIVE//PHYSICIAN REFERRAL
          hospital_admission_10:
            code: ADMISSION//EU OBSERVATION//PHYSICIAN REFERRAL
          hospital_admission_11:
            code: ADMISSION//OBSERVATION ADMIT//TRANSFER FROM HOSPITAL
          hospital_admission_12:
            code: ADMISSION//OBSERVATION ADMIT//WALK-IN/SELF REFERRAL
          hospital_admission_13:
            code: ADMISSION//DIRECT EMER.//CLINIC REFERRAL
          hospital_admission_14:
            code: ADMISSION//EU OBSERVATION//WALK-IN/SELF REFERRAL
          hospital_admission_15:
            code: ADMISSION//EW EMER.//TRANSFER FROM HOSPITAL
          hospital_admission_16:
            code: ADMISSION//EW EMER.//PHYSICIAN REFERRAL
          hospital_admission_17:
            code: ADMISSION//AMBULATORY OBSERVATION//PROCEDURE SITE
          hospital_admission_18:
            code: ADMISSION//URGENT//INTERNAL TRANSFER TO OR FROM PSYCH
          hospital_admission_19:
            code: ADMISSION//EW EMER.//PROCEDURE SITE
          hospital_admission_20:
            code: ADMISSION//EW EMER.//WALK-IN/SELF REFERRAL
          hospital_admission_21:
            code: ADMISSION//AMBULATORY OBSERVATION//PACU
          hospital_admission_22:
            code: ADMISSION//EW EMER.//PACU
          hospital_admission_23:
            code: ADMISSION//OBSERVATION ADMIT//CLINIC REFERRAL
          hospital_admission_24:
            code: ADMISSION//DIRECT OBSERVATION//TRANSFER FROM HOSPITAL
          hospital_admission_25:
            code: ADMISSION//URGENT//TRANSFER FROM SKILLED NURSING FACILITY
          hospital_admission_26:
            code: ADMISSION//EU OBSERVATION//TRANSFER FROM HOSPITAL
          hospital_admission_27:
            code: ADMISSION//DIRECT OBSERVATION//CLINIC REFERRAL
          hospital_admission_28:
            code: ADMISSION//OBSERVATION ADMIT//TRANSFER FROM SKILLED NURSING FACILITY
          hospital_admission_29:
            code: ADMISSION//DIRECT OBSERVATION//EMERGENCY ROOM
          hospital_admission_30:
            code: ADMISSION//DIRECT OBSERVATION//WALK-IN/SELF REFERRAL
          hospital_admission_31:
            code: ADMISSION//EU OBSERVATION//CLINIC REFERRAL
          hospital_admission_32:
            code: ADMISSION//EW EMER.//TRANSFER FROM SKILLED NURSING FACILITY
          hospital_admission_33:
            code: ADMISSION//EW EMER.//INTERNAL TRANSFER TO OR FROM PSYCH
          hospital_admission_34:
            code: ADMISSION//URGENT//WALK-IN/SELF REFERRAL
          hospital_admission_35:
            code: ADMISSION//DIRECT OBSERVATION//TRANSFER FROM SKILLED NURSING FACILITY
          hospital_admission_36:
            code: ADMISSION//EW EMER.//CLINIC REFERRAL
          hospital_admission_37:
            code: ADMISSION//OBSERVATION ADMIT//PROCEDURE SITE
          hospital_admission_38:
            code: ADMISSION//DIRECT OBSERVATION//PROCEDURE SITE
          hospital_admission_39:
            code: ADMISSION//EU OBSERVATION//TRANSFER FROM SKILLED NURSING FACILITY
          hospital_admission_40:
            code: ADMISSION//URGENT//CLINIC REFERRAL
          hospital_admission_41:
            code: ADMISSION//EW EMER.//INFORMATION NOT AVAILABLE
          hospital_admission_42:
            code: ADMISSION//URGENT//PROCEDURE SITE
          hospital_admission_43:
            code: ADMISSION//OBSERVATION ADMIT//INTERNAL TRANSFER TO OR FROM PSYCH
          hospital_admission_44:
            code: ADMISSION//URGENT//EMERGENCY ROOM
          hospital_admission_45:
            code: ADMISSION//DIRECT OBSERVATION//PACU
          hospital_admission_46:
            code: ADMISSION//OBSERVATION ADMIT//PACU
          hospital_admission_47:
            code: ADMISSION//MEDICAL
          hospital_admission_48:
            code: ADMISSION//OBSERVATION ADMIT//INFORMATION NOT AVAILABLE
          hospital_admission_49:
            code: ADMISSION//DIRECT OBSERVATION//INFORMATION NOT AVAILABLE
          hospital_admission_50:
            code: ADMISSION//URGENT//PACU
          hospital_admission_51:
            code: ADMISSION//EW EMER.//AMBULATORY SURGERY TRANSFER
          hospital_admission_52:
            code: ADMISSION//DIRECT OBSERVATION//INTERNAL TRANSFER TO OR FROM PSYCH
          hospital_admission_53:
            code: ADMISSION//URGENT//AMBULATORY SURGERY TRANSFER
          hospital_admission_54:
            code: ADMISSION//SURGICAL SAME DAY ADMISSION//TRANSFER FROM SKILLED NURSING FACILITY
          hospital_admission_55:
            code: ADMISSION//EU OBSERVATION//INFORMATION NOT AVAILABLE
          hospital_admission_56:
            code: ADMISSION//URGENT//INFORMATION NOT AVAILABLE
          hospital_admission_57:
            code: ADMISSION//EU OBSERVATION//AMBULATORY SURGERY TRANSFER
          hospital_admission_58:
            code: ADMISSION//SURGICAL SAME DAY ADMISSION//TRANSFER FROM HOSPITAL
          admission:
            expr: or(hospital_admission_0,hospital_admission_1,hospital_admission_2,hospital_admission_3,hospital_admission_4,hospital_admission_5,hospital_admission_6,hospital_admission_7,hospital_admission_8,hospital_admission_9,hospital_admission_10,hospital_admission_11,hospital_admission_12,hospital_admission_13,hospital_admission_14,hospital_admission_15,hospital_admission_16,hospital_admission_17,hospital_admission_18,hospital_admission_19,hospital_admission_20,hospital_admission_21,hospital_admission_22,hospital_admission_23,hospital_admission_24,hospital_admission_25,hospital_admission_26,hospital_admission_27,hospital_admission_28,hospital_admission_29,hospital_admission_30,hospital_admission_31,hospital_admission_32,hospital_admission_33,hospital_admission_34,hospital_admission_35,hospital_admission_36,hospital_admission_37,hospital_admission_38,hospital_admission_39,hospital_admission_40,hospital_admission_41,hospital_admission_42,hospital_admission_43,hospital_admission_44,hospital_admission_45,hospital_admission_46,hospital_admission_47,hospital_admission_48,hospital_admission_49,hospital_admission_50,hospital_admission_51,hospital_admission_52,hospital_admission_53,hospital_admission_54,hospital_admission_55,hospital_admission_56,hospital_admission_57,hospital_admission_58)

          hospital_discharge_0:
            code: DISCHARGE//HOME
          hospital_discharge_1:
            code: DISCHARGE//UNK
          hospital_discharge_2:
            code: DISCHARGE//HOME HEALTH CARE
          hospital_discharge_3:
            code: DISCHARGE//SKILLED NURSING FACILITY
          hospital_discharge_4:
            code: DISCHARGE//REHAB
          hospital_discharge_5:
            code: DISCHARGE//DIED
          hospital_discharge_6:
            code: DISCHARGE//CHRONIC/LONG TERM ACUTE CARE
          hospital_discharge_7:
            code: DISCHARGE//HOSPICE
          hospital_discharge_8:
            code: DISCHARGE//HOME_AMA
          hospital_discharge_9:
            code: DISCHARGE//PSYCH FACILITY
          hospital_discharge_10:
            code: DISCHARGE//ACUTE HOSPITAL
          hospital_discharge_11:
            code: DISCHARGE//OTHER FACILITY
          hospital_discharge_12:
            code: DISCHARGE//ASSISTED LIVING
          hospital_discharge_13:
            code: DISCHARGE//HEALTHCARE FACILITY
          discharge:
            expr: or(hospital_discharge_0,hospital_discharge_1,hospital_discharge_2,hospital_discharge_3,hospital_discharge_4,hospital_discharge_5,hospital_discharge_6,hospital_discharge_7,hospital_discharge_8,hospital_discharge_9,hospital_discharge_10,hospital_discharge_11,hospital_discharge_12,hospital_discharge_13)

          death:
            code: DEATH

          discharge_or_death:
            expr: or(discharge, death)

        trigger: discharge

        windows:
          data_within_5yr_of_admit:
            start: end - 1825d
            end: prior_admission.start
            start_inclusive: True
            end_inclusive: False
            has:
              _ANY_EVENT: (1, None)
          prior_admission:
            start: end <- admission
            end: trigger
            start_inclusive: True
            end_inclusive: False
            has:
              discharge_or_death: (None, 0)
          input:
            start: NULL
            end: trigger
            start_inclusive: True
            end_inclusive: True
            index_timestamp: end
          target:
            start: input.end
            end: start + 30d
            start_inclusive: False
            end_inclusive: True
            label: admission
          censor_protection:
            start: target.end
            end: null
            start_inclusive: False
            end_inclusive: True
            has:
              _ANY_EVENT: (1, None)
      """,  # noqa: E501
}

WANT_SHARDS = {
    "inhospital_mortality": parse_labels_yaml(
        """
  "0": |-2
    subject_id,prediction_time,boolean_value,integer_value,float_value,categorical_value
    1,01/15/2020 15:14,0,,,
    1,01/19/2022 04:46,0,,,
    1,01/25/2022 08:11,1,,,

  "1": |-2
    subject_id,prediction_time,boolean_value,integer_value,float_value,categorical_value
    3,03/19/2024 17:11,0,,,
    3,03/30/2024 11:00,0,,,
    """
    ),
    "HF_derived_readmission": parse_labels_yaml(
        """
  "0": |-2
    subject_id,prediction_time,boolean_value,integer_value,float_value,categorical_value
    1,01/20/2022 08:00,1,,,

  "1": |-2
    subject_id,prediction_time,boolean_value,integer_value,float_value,categorical_value
    """
    ),
    "nested_preds_readmission": parse_labels_yaml(
        """
  "0": |-2
    subject_id,prediction_time,boolean_value,integer_value,float_value,categorical_value
    1,01/20/2022 08:00,1,,,

  "1": |-2
    subject_id,prediction_time,boolean_value,integer_value,float_value,categorical_value
    3,01/20/2020 15:18,0,,,
    3,03/28/2024 10:00,1,,,
    3,04/19/2024 13:32,0,,,
    """
    ),
}


def test_meds():
    cli_test(
        input_files=MEDS_SHARDS,
        task_configs=TASKS,
        want_outputs_by_task=WANT_SHARDS,
        data_standard="meds",
    )



"""Main script for end-to-end cohort extraction."""

import sys
from importlib.resources import files

import hydra
import polars as pl
import pyarrow as pa
import pyarrow.parquet as pq
from loguru import logger
from meds import label_schema, prediction_time_field, subject_id_field
from omegaconf import DictConfig

config_yaml = files("aces").joinpath("configs/_aces.yaml")

if len(sys.argv) == 1:
    print("Usage: aces-cli [OPTIONS]")
    print("Try 'aces-cli --help' for more information.")
    print("For more information, visit: https://eventstreamaces.readthedocs.io/en/latest/usage.html")
    sys.exit(1)

MEDS_LABEL_MANDATORY_TYPES = {
    subject_id_field: pl.Int64,
}

MEDS_LABEL_OPTIONAL_TYPES = {
    "boolean_value": pl.Boolean,
    "integer_value": pl.Int64,
    "float_value": pl.Float64,
    "categorical_value": pl.String,
    prediction_time_field: pl.Datetime("us"),
}


def get_and_validate_label_schema(df: pl.DataFrame) -> pa.Table:
    """Validates the schema of a MEDS data DataFrame.

    This function validates the schema of a MEDS label DataFrame, ensuring that it has the correct columns
    and that the columns are of the correct type. This function will:
      1. Re-type any of the mandator MEDS column to the appropriate type.
      2. Attempt to add the ``numeric_value`` or ``time`` columns if either are missing, and set it to `None`.
         It will not attempt to add any other missing columns even if ``do_retype`` is `True` as the other
         columns cannot be set to `None`.

    Args:
        df: The MEDS label DataFrame to validate.

    Returns:
        pa.Table: The validated MEDS data DataFrame, with columns re-typed as needed.

    Raises:
        ValueError: if do_retype is False and the MEDS data DataFrame is not schema compliant.

    Examples:
        >>> df = pl.DataFrame({})
        >>> get_and_validate_label_schema(df) # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        ValueError: MEDS Data DataFrame must have a 'subject_id' column of type Int64.
                    MEDS Data DataFrame must have a 'prediction_time' column of type String.
                        Datetime(time_unit='us', time_zone=None).
        >>> from datetime import datetime
        >>> df = pl.DataFrame({
        ...     subject_id_field: pl.Series([1, 3, 2], dtype=pl.UInt32),
        ...     "time": [datetime(2021, 1, 1), datetime(2021, 1, 2), datetime(2021, 1, 3)],
        ...     "boolean_value": [1, 0, 100],
        ... })
        >>> get_and_validate_label_schema(df)
        pyarrow.Table
        subject_id: int64
        time: timestamp[us]
        boolean_value: bool
        integer_value: int64
        float_value: float
        categorical_value: string
        ----
        subject_id: [[1,3,2]]
        time: [[2021-01-01 00:00:00.000000,2021-01-02 00:00:00.000000,2021-01-03 00:00:00.000000]]
        boolean_value: [[true,false,true]]
        integer_value: [[null,null,null]]
        float_value: [[null,null,null]]
        categorical_value: [[null,null,null]]
    """

    schema = df.schema
    if "prediction_time" not in schema:
        logger.warning(
            "Output DataFrame is missing a 'prediction_time' column. If this is not intentional, add a "
            "'index_timestamp' (yes, it should be different) key to the task configuration identifying "
            "which window's start or end time to use as the prediction time."
        )

    errors = []
    for col, dtype in MEDS_LABEL_MANDATORY_TYPES.items():
        if col in schema and schema[col] != dtype:
            df = df.with_columns(pl.col(col).cast(dtype, strict=False))
        elif col not in schema:
            errors.append(f"MEDS Data DataFrame must have a '{col}' column of type {dtype}.")

    if errors:
        raise ValueError("\n".join(errors))

    for col, dtype in MEDS_LABEL_OPTIONAL_TYPES.items():
        if col in schema and schema[col] != dtype:
            df = df.with_columns(pl.col(col).cast(dtype, strict=False))
        elif col not in schema:
            df = df.with_columns(pl.lit(None, dtype=dtype).alias(col))

    extra_cols = [
        c for c in schema if c not in MEDS_LABEL_MANDATORY_TYPES and c not in MEDS_LABEL_OPTIONAL_TYPES
    ]
    if extra_cols:
        err_cols_str = "\n".join(f"  - {c}" for c in extra_cols)
        logger.warning(
            "Output contains columns that are not valid MEDS label columns. For now, we are dropping them.\n"
            "If you need these columns, please comment on https://github.com/justin13601/ACES/issues/97\n"
            f"Columns:\n{err_cols_str}"
        )
        df = df.drop(extra_cols)

    df = df.select(
        subject_id_field,
        "prediction_time",
        "boolean_value",
        "integer_value",
        "float_value",
        "categorical_value",
    )

    return df.to_arrow().cast(label_schema)


@hydra.main(version_base=None, config_path=str(config_yaml.parent.resolve()), config_name=config_yaml.stem)
def main(cfg: DictConfig):
    import os
    from datetime import datetime
    from pathlib import Path

    from hydra.core.hydra_config import HydraConfig
    from omegaconf import OmegaConf

    from . import config, predicates, query, utils

    utils.hydra_loguru_init(f"{HydraConfig.get().job.name}.log")

    st = datetime.now()

    # Set output path
    cohort_dir = Path(cfg.cohort_dir)
    cohort_dir.mkdir(exist_ok=True, parents=True)

    # load configuration
    logger.info(f"Loading config from '{cfg.config_path}'")
    if cfg.predicates_path:
        logger.info(f"Overriding predicates and/or demographics from '{cfg.predicates_path}'")
        predicates_path = Path(cfg.predicates_path)
    else:
        predicates_path = None
    task_cfg = config.TaskExtractorConfig.load(
        config_path=Path(cfg.config_path), predicates_path=predicates_path
    )

    logger.info(f"Attempting to get predicates dataframe given:\n{OmegaConf.to_yaml(cfg.data)}")
    predicates_df = predicates.get_predicates_df(task_cfg, cfg.data)

    # query results
    result = query.query(task_cfg, predicates_df)
    result_is_empty = len(result) == 0

    # save results to parquet
    os.makedirs(os.path.dirname(cfg.output_filepath), exist_ok=True)

    if cfg.data.standard.lower() == "meds":
        for in_col, out_col in [
            ("subject_id", subject_id_field),
            ("index_timestamp", "prediction_time"),
            ("label", "boolean_value"),
        ]:
            if in_col in result.columns:
                result = result.rename({in_col: out_col})
        if subject_id_field not in result.columns:
            if not result_is_empty:
                raise ValueError("Output dataframe is missing a 'subject_id' column.")
            else:
                logger.warning("Output dataframe is empty; adding an empty patient ID column.")
                result = result.with_columns(pl.lit(None, dtype=pl.Int64).alias(subject_id_field))
                result = result.head(0)
        if cfg.window_stats_dir:
            Path(cfg.window_stats_filepath).parent.mkdir(exist_ok=True, parents=True)
            result.write_parquet(cfg.window_stats_filepath)
        result = get_and_validate_label_schema(result)
        pq.write_table(result, cfg.output_filepath)
    else:
        result.write_parquet(cfg.output_filepath, use_pyarrow=True)
    logger.info(f"Completed in {datetime.now() - st}. Results saved to '{cfg.output_filepath}'.")


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import os
import re
import sys
from pathlib import Path


def expand_shards(*shards: str) -> str:
    """This function expands a set of shard prefixes and number of shards into a list of all shards or expands
    a directory into a list of all files within it.

    This can be useful with Hydra applications where you wish to expand a list of options for the sweeper to
    sweep over but can't use an OmegaConf resolver as those are evaluated after the sweep has been
    initialized.

    Args:
        shards: A list of shard prefixes and number of shards to expand, or a directory to list all files.

    Returns: A comma-separated list of all shards, expanded to the specified number, or all files in the
        directory.

    Examples:
        >>> import polars as pl
        >>> import tempfile

        >>> expand_shards("train/4", "val/IID/1", "val/prospective/1")
        'train/0,train/1,train/2,train/3,val/IID/0,val/prospective/0'
        >>> expand_shards("data/data_4", "data/test_4")
        'data/data_0,data/data_1,data/data_2,data/data_3,data/test_0,data/test_1,data/test_2,data/test_3'

        >>> parquet_data = pl.DataFrame({
        ...     "subject_id": [1, 1, 1, 2, 3],
        ...     "time": ["1/1/1989 00:00", "1/1/1989 01:00", "1/1/1989 01:00", "1/1/1989 02:00", None],
        ...     "code": ['admission', 'discharge', 'discharge', 'admission', "gender"],
        ... }).with_columns(pl.col("time").str.strptime(pl.Datetime, format="%m/%d/%Y %H:%M"))

        >>> with tempfile.TemporaryDirectory() as tmpdirname:
        ...     for i in range(4):
        ...         if i in (0, 2):
        ...             data_path = Path(tmpdirname) / f"evens/0/file_{i}.parquet"
        ...             data_path.parent.mkdir(parents=True, exist_ok=True)
        ...         else:
        ...             data_path = Path(tmpdirname) / f"{i}.parquet"
        ...         parquet_data.write_parquet(data_path)
        ...     json_fp = Path(tmpdirname) / "4.json"
        ...     _ = json_fp.write_text('["foo"]')
        ...     result = expand_shards(tmpdirname)
        ...     sorted(result.split(","))
        ['1', '3', 'evens/0/file_0', 'evens/0/file_2']
    """

    result = []
    for arg in shards:
        if os.path.isdir(arg):
            # If the argument is a directory, take all parquet files in any subdirs of the directory
            result.extend(
                str(x.relative_to(Path(arg)).with_suffix("")) for x in Path(arg).glob("**/*.parquet")
            )
        else:
            # Otherwise, treat it as a shard prefix and number of shards
            match = re.match(r"(.+)([/_])(\d+)$", arg)
            if match:
                prefix = match.group(1)
                delimiter = match.group(2)
                num = int(match.group(3))
                result.extend(f"{prefix}{delimiter}{i}" for i in range(num))
            else:
                raise ValueError(f"Invalid shard format: {arg}")

    return ",".join(result)


def main():
    print(expand_shards(*sys.argv[1:]))


if __name__ == "__main__":
    main()



import io
import os
import sys
from contextlib import contextmanager
from datetime import timedelta

import hydra
from bigtree import print_tree
from loguru import logger
from pytimeparse import parse


def parse_timedelta(time_str: str) -> timedelta:
    """Parse a time string and return a timedelta object.

    Using time expression parser: https://github.com/wroberts/pytimeparse

    Args:
        time_str: The time string to parse.

    Returns:
        datetime.timedelta: The parsed timedelta object.

    Examples:
        >>> parse_timedelta("1 days")
        datetime.timedelta(days=1)
        >>> parse_timedelta("1 day")
        datetime.timedelta(days=1)
        >>> parse_timedelta("1 days 2 hours 3 minutes 4 seconds")
        datetime.timedelta(days=1, seconds=7384)
        >>> parse_timedelta('1 day, 14:20:16')
        datetime.timedelta(days=1, seconds=51616)
        >>> parse_timedelta('365 days')
        datetime.timedelta(days=365)
    """
    if not time_str:
        return timedelta(days=0)

    return timedelta(seconds=parse(time_str))


@contextmanager
def capture_output():
    """A context manager to capture stdout output.

    This can eventually be eliminated if https://github.com/kayjan/bigtree/issues/285 is resolved.

    Examples:
        >>> with capture_output() as captured:
        ...     print("Hello, world!")
        >>> captured.getvalue().strip()
        'Hello, world!'
    """
    new_out = io.StringIO()  # Create a StringIO object to capture output
    old_out = sys.stdout  # Save the current stdout so we can restore it later
    try:
        sys.stdout = new_out  # Redirect stdout to our StringIO object
        yield new_out
    finally:
        sys.stdout = old_out  # Restore the original stdout


def log_tree(node):
    """Logs the tree structure using logging.info."""
    with capture_output() as captured:
        print_tree(node, style="const_bold")  # This will print to the captured StringIO instead of stdout
    logger.info("\n" + captured.getvalue())  # Log the captured output


def hydra_loguru_init(filename) -> None:
    """Must be called from a hydra main!"""
    hydra_path = hydra.core.hydra_config.HydraConfig.get().runtime.output_dir
    logger.add(os.path.join(hydra_path, filename))



"""This module contains the main function for querying a task.

It accepts the configuration file and predicate columns, builds the tree, and recursively queries the tree.
"""


import polars as pl
from bigtree import preorder_iter
from loguru import logger

from .config import TaskExtractorConfig
from .constraints import check_constraints, check_static_variables
from .extract_subtree import extract_subtree
from .utils import log_tree


def query(cfg: TaskExtractorConfig, predicates_df: pl.DataFrame) -> pl.DataFrame:
    """Query a task using the provided configuration file and predicates dataframe.

    Args:
        cfg: TaskExtractorConfig object of the configuration file.
        predicates_df: Polars predicates dataframe.

    Returns:
        polars.DataFrame: The result of the task query, containing subjects who satisfy the conditions
            defined in cfg. Timestamps for the start/end boundaries of each window specified in the task
            configuration, as well as predicate counts for each window, are provided.

    Raises:
        TypeError: If predicates_df is not a polars.DataFrame.
        ValueError: If the (subject_id, timestamp) columns are not unique.

    Examples:
    These examples are limited for now; see the `tests` directory for full examples.
        >>> cfg = None # This is obviously invalid, but we're just testing the error case.
        >>> predicates_df = {"subject_id": [1, 1], "timestamp": [1, 1]}
        >>> query(cfg, predicates_df)
        Traceback (most recent call last):
            ...
        TypeError: Predicates dataframe type must be a polars.DataFrame. Got: <class 'dict'>.
        >>> query(cfg, pl.DataFrame(predicates_df))
        Traceback (most recent call last):
            ...
        ValueError: The (subject_id, timestamp) columns must be unique.
        >>> from datetime import datetime
        >>> from .config import PlainPredicateConfig, WindowConfig, EventConfig
        >>> cfg = TaskExtractorConfig(
        ...     predicates={"A": PlainPredicateConfig("A")},
        ...     trigger=EventConfig("_ANY_EVENT"),
        ...     windows={
        ...         "pre": WindowConfig(None, "trigger", True, False),
        ...         "post": WindowConfig("pre.end", None, True, True),
        ...     },
        ... )
        >>> predicates_df = pl.DataFrame({
        ...     "subject_id": [1, 1, 3],
        ...     "timestamp": [datetime(1980, 12, 28), datetime(2010, 6, 20), datetime(2010, 5, 11)],
        ...     "A": [False, False, False],
        ...     "_ANY_EVENT": [True, True, True],
        ... })
        >>> query(cfg, predicates_df).select("subject_id", "trigger")
        shape: (3, 2)
        ┌────────────┬─────────────────────┐
        │ subject_id ┆ trigger             │
        │ ---        ┆ ---                 │
        │ i64        ┆ datetime[μs]        │
        ╞════════════╪═════════════════════╡
        │ 1          ┆ 1980-12-28 00:00:00 │
        │ 1          ┆ 2010-06-20 00:00:00 │
        │ 3          ┆ 2010-05-11 00:00:00 │
        └────────────┴─────────────────────┘
    """
    if not isinstance(predicates_df, pl.DataFrame):
        raise TypeError(f"Predicates dataframe type must be a polars.DataFrame. Got: {type(predicates_df)}.")

    logger.info("Checking if '(subject_id, timestamp)' columns are unique...")

    is_unique = predicates_df.n_unique(subset=["subject_id", "timestamp"]) == predicates_df.shape[0]

    if not is_unique:
        raise ValueError("The (subject_id, timestamp) columns must be unique.")

    log_tree(cfg.window_tree)

    logger.info("Beginning query...")

    static_variables = [pred for pred in cfg.predicates if cfg.predicates[pred].static]
    if static_variables:
        logger.info("Static variable criteria specified, filtering patient demographics...")
        predicates_df = check_static_variables(static_variables, predicates_df)
    else:
        logger.info("No static variable criteria specified, removing all rows with null timestamps...")
        predicates_df = predicates_df.drop_nulls(subset=["subject_id", "timestamp"])

    if predicates_df.is_empty():
        logger.warning("No valid rows found after filtering patient demographics. Exiting.")
        return pl.DataFrame()

    logger.info("Identifying possible trigger nodes based on the specified trigger event...")
    prospective_root_anchors = check_constraints({cfg.trigger.predicate: (1, None)}, predicates_df).select(
        "subject_id", pl.col("timestamp").alias("subtree_anchor_timestamp")
    )

    if prospective_root_anchors.is_empty():
        logger.warning(f"No valid rows found for the trigger event '{cfg.trigger.predicate}'. Exiting.")
        return pl.DataFrame()

    result = extract_subtree(cfg.window_tree, prospective_root_anchors, predicates_df)
    if result.is_empty():
        logger.info("No valid rows found.")
    else:
        # number of patients
        logger.info(
            f"Done. {result.shape[0]:,} valid rows returned corresponding to "
            f"{result['subject_id'].n_unique():,} subjects."
        )

    result = result.rename({"subtree_anchor_timestamp": "trigger"})

    to_return_cols = [
        "subject_id",
        "trigger",
        *[f"{node.node_name}_summary" for node in preorder_iter(cfg.window_tree)][1:],
    ]

    # add label column if specified
    if cfg.label_window:
        logger.info(
            f"Extracting label '{cfg.windows[cfg.label_window].label}' from window "
            f"'{cfg.label_window}'..."
        )
        label_col = "end" if cfg.windows[cfg.label_window].root_node == "start" else "start"
        result = result.with_columns(
            pl.col(f"{cfg.label_window}.{label_col}_summary")
            .struct.field(cfg.windows[cfg.label_window].label)
            .alias("label")
        )
        to_return_cols.insert(1, "label")

    # add index_timestamp column if specified
    if cfg.index_timestamp_window:
        logger.info(
            f"Setting index timestamp as '{cfg.windows[cfg.index_timestamp_window].index_timestamp}' "
            f"of window '{cfg.index_timestamp_window}'..."
        )
        index_timestamp_col = (
            "end" if cfg.windows[cfg.index_timestamp_window].root_node == "start" else "start"
        )
        result = result.with_columns(
            pl.col(f"{cfg.index_timestamp_window}.{index_timestamp_col}_summary")
            .struct.field(f"timestamp_at_{cfg.windows[cfg.index_timestamp_window].index_timestamp}")
            .alias("index_timestamp")
        )
        to_return_cols.insert(1, "index_timestamp")

    return result.select(to_return_cols)
